# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module that mints, stores, registers and resolves persistent identifiers.

Invenio-PIDStore is a core component of Invenio responsible for managing a
mapping from persistent identifiers to internal objects. The mapping is used
to resolve requests for e.g. an external integer record identifier to an
internal metadata record.

PIDStore consists of:

- `Persistent identifier API` - for working with persistent identifiers (
  create, reserve, register, delete, and redirect).
- `Resolver` - given a persistent identifier retrieve the assigned internal
  object.
- `Providers` - wrappers around a persistent identifier to provide extra
  functionality (e.g. interaction with remote services).
- `Minters` - small functions that are responsible for minting a specific
  persistent identifier type for a specific internal object type in as
  automatic way as possible.
- `Fetchers` - small functions that are responsible for returning a minted
  persistent identifier.

Initialization
--------------
First create a Flask application with a Click support (Flask version 0.11+):

>>> from flask import Flask
>>> app = Flask('myapp')
>>> app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite://'

You initialize PIDStore like a normal Flask extension, however Invenio-PIDStore
is dependent on Invenio-DB so you need to initialize both extensions:

>>> from invenio_db import InvenioDB
>>> from invenio_pidstore import InvenioPIDStore
>>> ext_db = InvenioDB(app)
>>> ext_pidstore = InvenioPIDStore(app)

In order for the following examples to work, you need to work within an
Flask application context so let's push one:

>>> ctx = app.app_context()
>>> ctx.push()

Also, for the examples to work we need to create the database and tables (note,
in this example we use an in-memory SQLite database):

>>> from invenio_db import db
>>> db.create_all()

Persistent identifiers
----------------------
Persistent identifiers map an external persistent identifier into an internal
object.

- A persistent identifier has **zero or one** assigned internal object.
- An internal object has **one or many** persistent identifiers.
- PIDStore can only map to internal objects which are **identified by UUIDs**.
- Persistent identifiers have a **type** (scheme) which is identified by a
  6 character string.
- All persistent identifier values are stored as string values, even if they
  may have a data type such as integer or UUID.

Persistent identifiers have a **state**:

- New: The persistent identifier is new.
- Reserved: The persistent identifier is reserved.
- Registered: The persistent identifier is registered.
- Redirected: The persistent identifier is redirected to another persistent
  identifier.
- Deleted: The persistent identifier is deleted/inactivated.

The different states might not all be relevant for different types of
identifiers. In addition persistent identifiers may be associated with a
specific **provider** (see below).

Creating PIDs
~~~~~~~~~~~~~
You can create a persistent identifier like this:

>>> from invenio_pidstore.models import PersistentIdentifier
>>> pid = PersistentIdentifier.create('doi', '10.1234/foo')

By default, persistent identifiers are created in the `new` state, and has
no object assigned:

>>> from invenio_pidstore.models import PIDStatus
>>> pid.status == PIDStatus.NEW
True
>>> pid.has_object()
False

Assign objects
~~~~~~~~~~~~~~
You can either assign objects after having created the persistent identifier:

>>> import uuid
>>> pid.assign('rec', uuid.uuid4())
True

or alternatively simply assign the object while you create the persistent
identifier (and also set the correct status immediately):

>>> pid = PersistentIdentifier.create('recid', '10',
...     object_type='rec', object_uuid=uuid.uuid4(),
...     status=PIDStatus.REGISTERED)
>>> pid.has_object()
True

Note, that once an object has been assigned you cannot assign a new object
unless you use the ``overwrite`` parameter:

>>> pid.assign('rec', uuid.uuid4())
Traceback (most recent call last):
...
invenio_pidstore.errors.PIDObjectAlreadyAssigned: ...
>>> pid.assign('rec', uuid.uuid4(), overwrite=True)
True

Modifying status
~~~~~~~~~~~~~~~~
You can modify the state of a persistent identifier:

>>> pid = PersistentIdentifier.create('doi', '10.1234/bar')
>>> pid.reserve()
True
>>> pid.register()
True
>>> pid.delete()
True

Note that if you delete a persistent identifier in state *new* it will be
completely removed from the database. Also, certain state changes are not
allowed (e.g. here trying to reserve a deleted persistent identifier):

>>> pid.reserve()
Traceback (most recent call last):
...
invenio_pidstore.errors.PIDInvalidAction: ...

Redirecting
~~~~~~~~~~~
You can redirect a registered  persistent identifier to another persistent
identifier. This can be useful in cases where you merge two underlying
records or you have multiple identifier schemes but one is preferred over
another.

>>> pid = PersistentIdentifier.create('recid', '11',
...     status=PIDStatus.REGISTERED)
>>> pid.redirect(PersistentIdentifier.get('recid', '10'))
True

This allows you easily retrieve the persistent identifier being redirected to:

>>> pid.get_redirect()
<PersistentIdentifier recid:10 / rec:... (R)>

Resolver
--------
The resolver is responsible for retrieving an internal object for a given
persistent identifier.

Following is an example for creating a record resolver.

>>> from invenio_pidstore.resolver import Resolver
>>> records = {}
>>> resolver = Resolver(pid_type='recid', object_type='rec',
...    getter=records.get)

Above creates a resolver that will translate ``recid`` persistent identifiers
to internal records. The getter argument must be a callable that takes one
argument which is the persistent identifiers object UUID.

Before we use the resolver, let's first create record and a persistent
identifier that we can use to test the resolver with:

>>> my_object_uuid = uuid.uuid4()
>>> records[my_object_uuid] = {'title': 'PIDStore test'}
>>> pid = PersistentIdentifier.create(
...     'recid', '12', object_type='rec', object_uuid=my_object_uuid,
...     status=PIDStatus.REGISTERED)

Using the resolver, we can now retrieve the record like this:

>>> pid, record = resolver.resolve('12')
>>> pid
<PersistentIdentifier recid:12 / rec:... (R)>
>>> record
{'title': 'PIDStore test'}

The resolver will only resolve **registered** persistent identifiers. Any
non-resolvable value will raise an exception. E.g. trying to resolve a reserved
persistent identifier will throw a
:py:exc:`invenio_pidstore.errors.PIDUnregistered` exception:

>>> pid = PersistentIdentifier.create(
...     'recid', '13', object_type='rec', object_uuid=my_object_uuid,
...     status=PIDStatus.RESERVED)
>>> resolver.resolve('13')
Traceback (most recent call last):
...
invenio_pidstore.errors.PIDUnregistered: ...

Providers
---------
Providers wrap the creation of persistent identifiers with extra functionality.
Use cases for this include automatically creating the persistent identifier or
retrieving the persistent identifier from an external service.

PIDStore comes by default with three providers:

- :py:class:`invenio_pidstore.providers.recordid_v2.RecordIdProviderV2` which
  creates random, checksummed, alphanumeric, PersistentIdentifier.
  (Recommended)
- :py:class:`invenio_pidstore.providers.datacite.DataCiteProvider` which
  creates and manages a PersistentIdentifier given a valid DOI.
- :py:class:`invenio_pidstore.providers.recordid.RecordIdProvider` which
  creates Invenio legacy integer record identifiers. (Not Recommended)

>>> from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2
>>> provider = RecordIdProviderV2.create()
>>> provider.pid.pid_type
'recid'
>>> provider.pid.pid_value  # doctest: +SKIP
'3sbk2-5j060'

Configure ``PIDSTORE_RECORDID_OPTIONS`` in ``config.py``, to construct a
custom `pid_value`.

Creating your own provider
~~~~~~~~~~~~~~~~~~~~~~~~~~
In order to create your own provider, you need to inherit from
:py:class:`invenio_pidstore.providers.base.BaseProvider` and override the
methods you want to customize.

Here's a minimal example:

>>> from invenio_pidstore.providers.base import BaseProvider
>>> class MyProvider(BaseProvider):
...     pid_type = 'my'
...     pid_provider = 'test'
...     default_status = PIDStatus.RESERVED

Minters
-------
Minters are small functions that are responsible for minting a specific
persistent identifier type for a specific internal object type in as automatic
way as possible.

A minter takes two arguments:

- an `object uuid` of the internal object which the persistent identifier
  should be assigned to.
- a `dictionary object`, which the minter may extract information from, and/or
  store extra information in. For instance, if a record object is passed, it
  might contain the DOI which a provider should mint, or alternatively, if the
  persistent identifier is generated, the minter can store the generated value
  into the record.

Below is an example minter which mints external UUIDs (not to be confused with
the internal object UUIDs), and stores the generated UUID into the dictionary
object (which could be e.g. a record):

>>> import uuid
>>> def my_uuid_minter(object_uuid, data):
...     pid = PersistentIdentifier.create('myuuid', str(uuid.uuid4()))
...     data['uuid'] = str(pid.object_uuid)
...     return pid

Registering minters
~~~~~~~~~~~~~~~~~~~
Minters are usually used by other modules and thus registered on the
Flask application.

First import the proxy to the current application's PIDStore:

>>> from invenio_pidstore import current_pidstore

Next, register a minter:

>>> current_pidstore.register_minter('my_uuid_minter', my_uuid_minter)

You can retrieve registered minters from the extension as well:

>>> current_pidstore.minters['my_uuid_minter']
<function my_uuid_minter at ...>


Fetchers
--------
Fetchers are small functions that are responsible for returning a minted
persistent identifier.

A fetcher takes two arguments:

- an `object uuid` of the internal object which the persistent identifier
  should be assigned to.
- a `dictionary object`, which the fetcher may extract information from. For
  instance, if a record object is passed, it might contain a previously minted
  DOI.

Below is an example fetcher which extracts external UUIDs previously minted in
the dictionary object (which could be e.g. a record):

>>> import uuid
>>> def my_uuid_fetcher(object_uuid, data):
...     return data['uuid']

Registering fetchers
~~~~~~~~~~~~~~~~~~~~
Fetchers are usually used by other modules and thus registered on the
Flask application.

First import the proxy to the current application's PIDStore:

>>> from invenio_pidstore import current_pidstore

Next, register a fetcher:

>>> current_pidstore.register_fetcher('my_uuid_fetcher', my_uuid_fetcher)

You can retrieve registered fetchers from the extension as well:

>>> current_pidstore.fetchers['my_uuid_fetcher']
<function my_uuid_fetcher at ...>


Entry points loading
~~~~~~~~~~~~~~~~~~~~
PIDStore will automatically register minters and fetchers defined by the entry
point groups ``invenio_pidstore.minters`` and . ``invenio_pidstore.fetchers``.

Example code:

.. code-block:: python

   # setup.py
   setup(
       # ...
       entry_points={
           'invenio_pidstore.minters': [
               'recid = invenio_pidstore.minters:recid_minter',
           ],
           'invenio_pidstore.fetchers': [
               'recid = invenio_pidstore.fetchers:recid_fetcher',
           ]})

Above is equivalent to:

.. code-block:: python

   from invenio_pidstore.minters import recid_minter
   from invenio_pidstore.fetchers import recid_fetcher
   current_pidstore.register_minter('recid', recid_minter)
   current_pidstore.register_fetcher('recid', recid_fetcher)

"""

from __future__ import absolute_import, print_function

from .ext import InvenioPIDStore
from .proxies import current_pidstore

__version__ = "2.0.0"

__all__ = ("__version__", "InvenioPIDStore", "current_pidstore")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Minimal Flask application example for development.

SPHINX-START

First install Invenio-PIDStore, setup the application and load
fixture data by running:

.. code-block:: console

   $ pip install -e .[all]
   $ cd examples
   $ ./app-setup.sh
   $ ./app-fixtures.sh

Next, start the development server:

.. code-block:: console

   $ export FLASK_APP=app.py FLASK_DEBUG=1
   $ flask run

Open the admin page:

.. code-block:: console

    $ open http://127.0.0.1:5000/admin/recordmetadata/

Login with:

    username: admin@inveniosoftware.org

    password: 123456

To reset the example application run:

.. code-block:: console

    $ ./app-teardown.sh

SPHINX-END

"""

from __future__ import absolute_import, print_function

import os

from flask import Flask
from flask_babelex import Babel
from flask_menu import Menu
from invenio_access import InvenioAccess
from invenio_accounts import InvenioAccounts
from invenio_accounts.views import blueprint as accounts_blueprint
from invenio_admin import InvenioAdmin
from invenio_db import InvenioDB
from invenio_records import InvenioRecords

from invenio_pidstore import InvenioPIDStore

# Create Flask application
app = Flask(__name__)
app.config.update(
    DB_VERSIONING_USER_MODEL=None,
    SECRET_KEY="test_key",
    SECURITY_PASSWORD_HASH="pbkdf2_sha512",
    SECURITY_PASSWORD_SALT="CHANGE_ME_ALSO",
    SECURITY_PASSWORD_SCHEMES=[
        "pbkdf2_sha512",
        "sha512_crypt",
        "invenio_aes_encrypted_email",
    ],
    SQLALCHEMY_DATABASE_URI=os.environ.get(
        "SQLALCHEMY_DATABASE_URI", "sqlite:///test.db"
    ),
    WTF_CSRF_ENABLED=False,
)

Babel(app)
Menu(app)
InvenioDB(app)
admin = InvenioAdmin(app)
InvenioPIDStore(app)
accounts = InvenioAccounts(app)
InvenioAccess(app)
InvenioRecords(app)

# register blueprints
app.register_blueprint(accounts_blueprint)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module storing implementations of PID providers."""

from __future__ import absolute_import, print_function

from ..models import PersistentIdentifier, PIDStatus


class BaseProvider(object):
    """Abstract class for persistent identifier provider classes."""

    pid_type = None
    """Default persistent identifier type."""

    pid_provider = None
    """Persistent identifier provider name."""

    default_status = PIDStatus.NEW
    """Default status for newly created PIDs by this provider."""

    @classmethod
    def create(
        cls,
        pid_type=None,
        pid_value=None,
        object_type=None,
        object_uuid=None,
        status=None,
        **kwargs
    ):
        """Create a new instance for the given type and pid.

        :param pid_type: Persistent identifier type. (Default: None).
        :param pid_value: Persistent identifier value. (Default: None).
        :param status: Current PID status.
            (Default: :attr:`invenio_pidstore.models.PIDStatus.NEW`)
        :param object_type: The object type is a string that identify its type.
            (Default: None).
        :param object_uuid: The object UUID. (Default: None).
        :returns: A :class:`invenio_pidstore.providers.base.BaseProvider`
            instance.
        """
        assert pid_value
        assert pid_type or cls.pid_type

        pid = PersistentIdentifier.create(
            pid_type or cls.pid_type,
            pid_value,
            pid_provider=cls.pid_provider,
            object_type=object_type,
            object_uuid=object_uuid,
            status=status or cls.default_status,
        )
        return cls(pid, **kwargs)

    @classmethod
    def get(cls, pid_value, pid_type=None, **kwargs):
        """Get a persistent identifier for this provider.

        :param pid_type: Persistent identifier type. (Default: configured
            :attr:`invenio_pidstore.providers.base.BaseProvider.pid_type`)
        :param pid_value: Persistent identifier value.
        :param kwargs: See
            :meth:`invenio_pidstore.providers.base.BaseProvider` required
            initialization properties.
        :returns: A :class:`invenio_pidstore.providers.base.BaseProvider`
            instance.
        """
        return cls(
            PersistentIdentifier.get(
                pid_type or cls.pid_type, pid_value, pid_provider=cls.pid_provider
            ),
            **kwargs
        )

    def __init__(self, pid, **kwargs):
        """Initialize provider using persistent identifier.

        :param pid: A :class:`invenio_pidstore.models.PersistentIdentifier`
            instance.
        """
        self.pid = pid
        assert pid.pid_provider == self.pid_provider

    def reserve(self):
        """Reserve a persistent identifier.

        This might or might not be useful depending on the service of the
        provider.

        See: :meth:`invenio_pidstore.models.PersistentIdentifier.reserve`.
        """
        return self.pid.reserve()

    def register(self):
        """Register a persistent identifier.

        See: :meth:`invenio_pidstore.models.PersistentIdentifier.register`.
        """
        return self.pid.register()

    def update(self):
        """Update information about the persistent identifier."""
        pass

    def delete(self):
        """Delete a persistent identifier.

        See: :meth:`invenio_pidstore.models.PersistentIdentifier.delete`.
        """
        return self.pid.delete()

    def sync_status(self):
        """Synchronize PIDstatus with remote service provider."""
        pass



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

from __future__ import absolute_import, print_function

import os
import shutil
import tempfile

import pytest
from flask import Flask
from invenio_db import InvenioDB
from invenio_i18n import InvenioI18N
from sqlalchemy_utils.functions import create_database, database_exists

from invenio_pidstore import InvenioPIDStore


@pytest.yield_fixture()
def app(request):
    """Flask application fixture."""
    # Set temporary instance path for sqlite
    instance_path = tempfile.mkdtemp()
    app = Flask("testapp", instance_path=instance_path)
    InvenioDB(app)
    InvenioI18N(app)
    InvenioPIDStore(app)

    app.config.update(
        SQLALCHEMY_DATABASE_URI=os.environ.get(
            "SQLALCHEMY_DATABASE_URI", "sqlite:///test.db"
        ),
        TESTING=True,
    )

    with app.app_context():
        yield app

    # Teardown instance path.
    shutil.rmtree(instance_path)


@pytest.yield_fixture()
def db(app):
    """Database fixture."""
    from invenio_db import db as db_

    if not database_exists(str(db_.engine.url)):
        create_database(str(db_.engine.url))
    db_.create_all()

    yield db_

    db_.session.remove()
    db_.drop_all()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Model tests."""

from __future__ import absolute_import, print_function

import uuid

import pytest
from mock import patch
from sqlalchemy.exc import SQLAlchemyError

from invenio_pidstore.errors import (
    PIDAlreadyExists,
    PIDDoesNotExistError,
    PIDInvalidAction,
    PIDObjectAlreadyAssigned,
)
from invenio_pidstore.models import PersistentIdentifier, PIDStatus, Redirect


@patch("invenio_pidstore.models.logger")
def test_pid_creation(logger, app, db):
    """Test pid creation."""
    with app.app_context():
        assert PersistentIdentifier.query.count() == 0
        pid = PersistentIdentifier.create("doi", "10.1234/foo")
        assert PersistentIdentifier.query.count() == 1
        assert pid.pid_type == "doi"
        assert pid.pid_value == "10.1234/foo"
        assert pid.pid_provider is None
        assert pid.status == PIDStatus.NEW
        assert pid.object_type is None
        assert pid.object_uuid is None
        assert logger.info.called

        rec_uuid = uuid.uuid4()
        pid = PersistentIdentifier.create(
            "rec",
            "2",
            status=PIDStatus.REGISTERED,
            object_type="rec",
            object_uuid=rec_uuid,
        )
        assert PersistentIdentifier.query.count() == 2
        assert pid.pid_type == "rec"
        assert pid.pid_value == "2"
        assert pid.pid_provider is None
        assert pid.status == PIDStatus.REGISTERED
        assert pid.object_type == "rec"
        assert pid.object_uuid == rec_uuid

        # Can't duplicate existing persistent identifier
        assert not logger.exception.called
        pytest.raises(PIDAlreadyExists, PersistentIdentifier.create, "rec", "2")
        assert logger.exception.called

        with patch("invenio_pidstore.models.db.session.begin_nested") as mock:
            mock.side_effect = SQLAlchemyError()
            pytest.raises(SQLAlchemyError, PersistentIdentifier.create, "rec", "2")
            assert logger.exception.call_args[0][0].startswith("Failed to create")


def test_alembic(app, db):
    """Test alembic recipes."""
    ext = app.extensions["invenio-db"]

    if db.engine.name == "sqlite":
        raise pytest.skip("Upgrades are not supported on SQLite.")

    assert not ext.alembic.compare_metadata()
    db.drop_all()
    ext.alembic.upgrade()

    assert not ext.alembic.compare_metadata()
    ext.alembic.stamp()
    ext.alembic.downgrade(target="96e796392533")
    ext.alembic.upgrade()

    assert not ext.alembic.compare_metadata()


def test_pidstatus_as():
    """Test PID status."""
    assert PIDStatus.NEW.title == "New"
    assert PIDStatus.RESERVED.title == "Reserved"
    assert next(iter(PIDStatus)) == "N"


def test_pid_get(app, db):
    """Test pid retrieval."""
    with app.app_context():
        PersistentIdentifier.create("doi", "10.1234/foo")
        assert PersistentIdentifier.get("doi", "10.1234/foo")
        pytest.raises(
            PIDDoesNotExistError, PersistentIdentifier.get, "doi", "10.1234/bar"
        )

        # PID with provider
        doi = "10.1234/a"
        PersistentIdentifier.create("doi", doi, pid_provider="dcite")
        assert PersistentIdentifier.get("doi", doi)
        assert PersistentIdentifier.get("doi", doi, pid_provider="dcite")
        pytest.raises(
            PIDDoesNotExistError,
            PersistentIdentifier.get,
            "doi",
            doi,
            pid_provider="cref",
        )

        # Retrieve by object
        myuuid = uuid.uuid4()
        doi = "10.1234/b"
        PersistentIdentifier.create("doi", doi, object_type="rec", object_uuid=myuuid)
        pid = PersistentIdentifier.get_by_object("doi", "rec", myuuid)
        assert pid.pid_value == doi

        pytest.raises(
            PIDDoesNotExistError,
            PersistentIdentifier.get_by_object,
            "doi",
            "rec",
            uuid.uuid4(),
        )


@patch("invenio_pidstore.models.logger")
def test_pid_assign(logger, app, db):
    """Test pid object assignment."""
    with app.app_context():
        # No assigned object
        pid = PersistentIdentifier.create("doi", "10.1234/foo")
        assert not pid.has_object()
        assert pid.get_assigned_object() is None
        assert pid.get_assigned_object("rec") is None

        # Assign object
        rec_uuid = uuid.uuid4()
        pid.assign("rec", rec_uuid)
        assert logger.info.call_args[0][0].startswith("Assigned")
        assert "pid" in logger.info.call_args[1]["extra"]
        assert pid.has_object()
        assert pid.get_assigned_object() == rec_uuid
        assert pid.get_assigned_object("rec") == rec_uuid
        assert pid.get_assigned_object("oth") is None
        # Doesnt' raise
        pid.assign("rec", rec_uuid)

        # Assign without overwrite (uuid as str and uuid)
        new_uuid = uuid.uuid4()
        pytest.raises(PIDObjectAlreadyAssigned, pid.assign, "rec", new_uuid)
        pytest.raises(PIDObjectAlreadyAssigned, pid.assign, "rec", str(new_uuid))

        # Assign with overwrite
        pid.assign("rec", str(new_uuid), overwrite=True)
        assert pid.has_object()
        assert pid.get_assigned_object() == new_uuid
        assert pid.get_assigned_object("rec") == new_uuid
        assert pid.get_assigned_object("oth") is None

        # Assign with SQLError
        pid = PersistentIdentifier.create("recid", "101")
        with patch("invenio_pidstore.models.db.session.begin_nested") as mock:
            mock.side_effect = SQLAlchemyError()
            pytest.raises(SQLAlchemyError, pid.assign, "rec", uuid.uuid4())


@patch("invenio_pidstore.models.logger")
def test_pid_unassign_noobject(logger, app, db):
    """Test unassign."""
    with app.app_context():
        pid = PersistentIdentifier.create("recid", "101")
        assert pid.unassign()
        pid.assign("rec", uuid.uuid4())
        with patch("invenio_pidstore.models.db.session.begin_nested") as mock:
            mock.side_effect = SQLAlchemyError()
            pytest.raises(SQLAlchemyError, pid.unassign)
            assert logger.exception.call_args[0][0].startswith("Failed to unassign")
            assert "pid" in logger.exception.call_args[1]["extra"]


def test_pid_assign_deleted(app, db):
    """Test pid object assignment."""
    with app.app_context():
        pid = PersistentIdentifier.create(
            "doi", "10.1234/foo", status=PIDStatus.DELETED
        )
        pytest.raises(PIDInvalidAction, pid.assign, "rec", uuid.uuid4())


@patch("invenio_pidstore.models.logger")
def test_reserve(logger, app, db):
    """Test pid reserve."""
    with app.app_context():
        i = 1
        for s in [PIDStatus.NEW, PIDStatus.RESERVED]:
            pid = PersistentIdentifier.create("rec", str(i), status=s)
            i += 1
            assert pid.reserve()
            assert logger.info.call_args[0][0].startswith("Reserved PID")
        for s in [PIDStatus.REGISTERED, PIDStatus.DELETED, PIDStatus.REDIRECTED]:
            pid = PersistentIdentifier.create("rec", str(i), status=s)
            i += 1
            pytest.raises(PIDInvalidAction, pid.reserve)

        # Test logging of bad errors.
        pid = PersistentIdentifier.create("rec", str(i))
        with patch("invenio_pidstore.models.db.session.begin_nested") as mock:
            mock.side_effect = SQLAlchemyError()
            pytest.raises(SQLAlchemyError, pid.reserve)
            assert logger.exception.call_args[0][0].startswith("Failed to reserve")
            assert "pid" in logger.exception.call_args[1]["extra"]


@patch("invenio_pidstore.models.logger")
def test_register(logger, app, db):
    """Test pid register."""
    with app.app_context():
        i = 1
        for s in [PIDStatus.NEW, PIDStatus.RESERVED]:
            pid = PersistentIdentifier.create("rec", str(i), status=s)
            i += 1
            assert pid.register()
            assert logger.info.call_args[0][0].startswith("Registered PID")
        for s in [PIDStatus.REGISTERED, PIDStatus.DELETED, PIDStatus.REDIRECTED]:
            pid = PersistentIdentifier.create("rec", str(i), status=s)
            i += 1
            pytest.raises(PIDInvalidAction, pid.register)

        # Test logging of bad errors.
        pid = PersistentIdentifier.create("rec", str(i), status=PIDStatus.RESERVED)
        with patch("invenio_pidstore.models.db.session.begin_nested") as mock:
            mock.side_effect = SQLAlchemyError()
            pytest.raises(SQLAlchemyError, pid.register)
            assert logger.exception.call_args[0][0].startswith("Failed to register")
            assert "pid" in logger.exception.call_args[1]["extra"]


@patch("invenio_pidstore.models.logger")
def test_delete(logger, app, db):
    """Test pid delete."""
    with app.app_context():
        i = 1
        for s in [
            PIDStatus.RESERVED,
            PIDStatus.RESERVED,
            PIDStatus.REDIRECTED,
            PIDStatus.DELETED,
        ]:
            pid = PersistentIdentifier.create("rec", str(i), status=s)
            i += 1
            assert pid.delete()
            assert logger.info.call_args[0][0] == "Deleted PID."

        # New persistent identifiers are removed completely
        count = PersistentIdentifier.query.count()
        pid = PersistentIdentifier.create("rec", str(i), status=PIDStatus.NEW)
        db.session.commit()
        assert PersistentIdentifier.query.count() == count + 1
        pid.delete()
        assert PersistentIdentifier.query.count() == count
        assert logger.info.call_args[0][0] == "Deleted PID (removed)."

        pid = PersistentIdentifier.create("rec", str(i + 1))
        with patch("invenio_pidstore.models.db.session.begin_nested") as mock:
            mock.side_effect = SQLAlchemyError()
            pytest.raises(SQLAlchemyError, pid.delete)
            assert logger.exception.call_args[0][0].startswith("Failed to delete")
            assert "pid" in logger.exception.call_args[1]["extra"]


@patch("invenio_pidstore.models.logger")
def test_redirect(logger, app, db):
    """Test redirection."""
    with app.app_context():
        pid1 = PersistentIdentifier.create(
            "rec",
            "1",
            status=PIDStatus.REGISTERED,
            object_type="rec",
            object_uuid=uuid.uuid4(),
        )
        pid2 = PersistentIdentifier.create(
            "doi",
            "2",
            status=PIDStatus.REGISTERED,
            object_type="rec",
            object_uuid=uuid.uuid4(),
        )

        # Can't redirect these statuses
        i = 10
        for s in [
            PIDStatus.NEW,
            PIDStatus.RESERVED,
            PIDStatus.DELETED,
        ]:
            pid = PersistentIdentifier.create("rec", str(i), status=s)
            i += 1
            pytest.raises(PIDInvalidAction, pid.redirect, pid1)

        pid = PersistentIdentifier.create("rec", str(i), status=PIDStatus.REGISTERED)

        # Can't redirect to non-exsting pid.
        pytest.raises(PIDDoesNotExistError, pid.redirect, PersistentIdentifier())

        pid.redirect(pid1)
        assert logger.info.call_args[0][0].startswith("Redirected")
        assert "pid" in logger.info.call_args[1]["extra"]
        assert pid.status == PIDStatus.REDIRECTED
        assert pid.object_type is None
        assert pid.object_uuid is not None
        new_pid = pid.get_redirect()
        assert new_pid.pid_type == "rec"
        assert new_pid.pid_value == "1"

        # You can redirect an already redirected pid
        pid.redirect(pid2)
        new_pid = pid.get_redirect()
        assert new_pid.pid_type == "doi"
        assert new_pid.pid_value == "2"

        # Assign with SQLError
        with patch("invenio_pidstore.models.db.session.begin_nested") as mock:
            mock.side_effect = SQLAlchemyError()
            pytest.raises(SQLAlchemyError, pid.redirect, "1")
            assert logger.exception.call_args[0][0].startswith("Failed to redirect")
            assert "pid" in logger.exception.call_args[1]["extra"]


def test_redirect_cleanup(app, db):
    """Test proper clean up from redirects."""
    with app.app_context():
        pid1 = PersistentIdentifier.create(
            "recid",
            "1",
            status=PIDStatus.REGISTERED,
            object_type="rec",
            object_uuid=uuid.uuid4(),
        )
        pid2 = PersistentIdentifier.create(
            "recid",
            "2",
            status=PIDStatus.REGISTERED,
            object_type="rec",
            object_uuid=uuid.uuid4(),
        )
        pid3 = PersistentIdentifier.create("recid", "3", status=PIDStatus.REGISTERED)
        db.session.commit()

        assert Redirect.query.count() == 0
        pid3.redirect(pid1)
        assert Redirect.query.count() == 1
        pid3.redirect(pid2)
        assert Redirect.query.count() == 1
        pytest.raises(PIDObjectAlreadyAssigned, pid3.assign, "rec", uuid.uuid4())
        pid3.unassign()
        assert Redirect.query.count() == 0


@patch("invenio_pidstore.models.logger")
def test_sync_status(logger, app, db):
    """Test sync status."""
    with app.app_context():
        pid = PersistentIdentifier.create(
            "rec",
            "1",
            status=PIDStatus.REGISTERED,
            object_type="rec",
            object_uuid=uuid.uuid4(),
        )
        pytest.raises(PIDInvalidAction, pid.reserve)
        calls = logger.info.call_count
        assert pid.sync_status(PIDStatus.NEW)
        assert logger.info.call_count == calls + 1
        assert pid.reserve()
        calls = logger.info.call_count
        assert pid.sync_status(PIDStatus.RESERVED)
        assert logger.info.call_count == calls

        with patch("invenio_pidstore.models.db.session.begin_nested") as mock:
            mock.side_effect = SQLAlchemyError()
            pytest.raises(SQLAlchemyError, pid.sync_status, PIDStatus.NEW)
            assert logger.exception.call_args[0][0].startswith("Failed to sync status")
            assert "pid" in logger.exception.call_args[1]["extra"]


def test_repr(app, db):
    """Test representation."""
    with app.app_context():
        pid = PersistentIdentifier.create(
            "recid",
            "1",
            status=PIDStatus.REGISTERED,
            object_type="rec",
            object_uuid="de3bb351-bc1a-4e51-8605-c6cd9589a560",
        )
        assert (
            str(pid) == "<PersistentIdentifier recid:1 / "
            "rec:de3bb351-bc1a-4e51-8605-c6cd9589a560 (R)>"
        )
        pid = PersistentIdentifier.create("recid", "2", status=PIDStatus.REGISTERED)
        assert str(pid) == "<PersistentIdentifier recid:2 (R)>"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio-PIDStore configuration."""

PIDSTORE_RECID_FIELD = "control_number"
"""Default record id field inside the json data.

This name will be used by the fetcher, to retrieve the record ID value from the
JSON, and by the minter, to store it inside the JSON.
"""


PIDSTORE_DATACITE_DOI_PREFIX = ""
"""Provide a DOI prefix here."""

PIDSTORE_RECORDID_OPTIONS = {"length": 10, "split_every": 5, "checksum": True}



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Click command-line interface for PIDStore management."""

from __future__ import absolute_import, print_function

import click
from flask.cli import with_appcontext
from invenio_db import db

from .proxies import current_pidstore


def process_status(ctx, param, value):
    """Return status value."""
    from .models import PIDStatus

    # Allow empty status
    if value is None:
        return None

    if not hasattr(PIDStatus, value):
        raise click.BadParameter(
            "Status needs to be one of {0}.".format(
                ", ".join([s.name for s in PIDStatus])
            )
        )
    return getattr(PIDStatus, value)


#
# PIDStore management commands
#


@click.group()
def pid():
    """PID-Store management commands."""


@pid.command()
@click.argument("pid_type")
@click.argument("pid_value")
@click.option("-s", "--status", default="NEW", callback=process_status)
@click.option("-t", "--type", "object_type", default=None)
@click.option("-i", "--uuid", "object_uuid", default=None)
@with_appcontext
def create(pid_type, pid_value, status, object_type, object_uuid):
    """Create new persistent identifier."""
    from .models import PersistentIdentifier

    if bool(object_type) ^ bool(object_uuid):
        raise click.BadParameter("Speficy both or any of --type and --uuid.")

    new_pid = PersistentIdentifier.create(
        pid_type,
        pid_value,
        status=status,
        object_type=object_type,
        object_uuid=object_uuid,
    )
    db.session.commit()
    click.echo("{0.pid_type} {0.pid_value} {0.pid_provider}".format(new_pid))


@pid.command()
@click.argument("pid_type")
@click.argument("pid_value")
@click.option("-s", "--status", default=None, callback=process_status)
@click.option("-t", "--type", "object_type", required=True)
@click.option("-i", "--uuid", "object_uuid", required=True)
@click.option("--overwrite", is_flag=True, default=False)
@with_appcontext
def assign(pid_type, pid_value, status, object_type, object_uuid, overwrite):
    """Assign persistent identifier."""
    from .models import PersistentIdentifier

    obj = PersistentIdentifier.get(pid_type, pid_value)
    if status is not None:
        obj.status = status
    obj.assign(object_type, object_uuid, overwrite=overwrite)
    db.session.commit()
    click.echo(obj.status)


@pid.command()
@click.argument("pid_type")
@click.argument("pid_value")
@with_appcontext
def unassign(pid_type, pid_value):
    """Unassign persistent identifier."""
    from .models import PersistentIdentifier

    obj = PersistentIdentifier.get(pid_type, pid_value)
    obj.unassign()
    db.session.commit()
    click.echo(obj.status)


@pid.command("get")
@click.argument("pid_type")
@click.argument("pid_value")
@with_appcontext
def get_object(pid_type, pid_value):
    """Get an object behind persistent identifier."""
    from .models import PersistentIdentifier

    obj = PersistentIdentifier.get(pid_type, pid_value)
    if obj.has_object():
        click.echo("{0.object_type} {0.object_uuid} {0.status}".format(obj))


@pid.command("dereference")
@click.argument("object_type")
@click.argument("object_uuid")
@click.option("-s", "--status", default=None, callback=process_status)
@with_appcontext
def dereference_object(object_type, object_uuid, status):
    """Show linked persistent identifier(s)."""
    from .models import PersistentIdentifier

    pids = PersistentIdentifier.query.filter_by(
        object_type=object_type, object_uuid=object_uuid
    )
    if status:
        pids = pids.filter_by(status=status)

    for found_pid in pids.all():
        click.echo("{0.pid_type} {0.pid_value} {0.pid_provider}".format(found_pid))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2022 RERO.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that stores and registers persistent identifiers."""

from __future__ import absolute_import, print_function

import importlib_metadata
import importlib_resources

from . import config
from .cli import pid as cmd
from .errors import PIDDoesNotExistError
from .models import PersistentIdentifier, logger


def pid_exists(value, pidtype=None):
    """Check if a persistent identifier exists.

    :param value: The PID value.
    :param pidtype: The pid value (Default: None).
    :returns: `True` if the PID exists.
    """
    try:
        PersistentIdentifier.get(pidtype, value)
        return True
    except PIDDoesNotExistError:
        return False


class _PIDStoreState(object):
    """Persistent identifier store state."""

    def __init__(
        self, app, minters_entry_point_group=None, fetchers_entry_point_group=None
    ):
        """Initialize state."""
        self.app = app
        self.minters = {}
        self.fetchers = {}
        if minters_entry_point_group:
            self.load_minters_entry_point_group(minters_entry_point_group)
        if fetchers_entry_point_group:
            self.load_fetchers_entry_point_group(fetchers_entry_point_group)

    def register_minter(self, name, minter):
        """Register a minter.

        :param name: Minter name.
        :param minter: The new minter.
        """
        if name not in self.minters:
            self.minters[name] = minter

    def register_fetcher(self, name, fetcher):
        """Register a fetcher.

        :param name: Fetcher name.
        :param fetcher: The new fetcher.
        """
        if name not in self.fetchers:
            self.fetchers[name] = fetcher

    def load_minters_entry_point_group(self, entry_point_group):
        """Load minters from an entry point group.

        :param entry_point_group: The entrypoint group.
        """
        for ep in importlib_metadata.entry_points(group=entry_point_group):
            self.register_minter(ep.name, ep.load())

    def load_fetchers_entry_point_group(self, entry_point_group):
        """Load fetchers from an entry point group.

        :param entry_point_group: The entrypoint group.
        """
        for ep in importlib_metadata.entry_points(group=entry_point_group):
            self.register_fetcher(ep.name, ep.load())


class InvenioPIDStore(object):
    """Invenio-PIDStore extension."""

    def __init__(
        self,
        app=None,
        minters_entry_point_group="invenio_pidstore.minters",
        fetchers_entry_point_group="invenio_pidstore.fetchers",
    ):
        """Extension initialization.

        :param minters_entry_point_group: The entrypoint for minters.
            (Default: `invenio_pidstore.minters`).
        :param fetchers_entry_point_group: The entrypoint for fetchers.
            (Default: `invenio_pidstore.fetchers`).
        """
        if app:
            self._state = self.init_app(
                app,
                minters_entry_point_group=minters_entry_point_group,
                fetchers_entry_point_group=fetchers_entry_point_group,
            )

    def init_app(
        self, app, minters_entry_point_group=None, fetchers_entry_point_group=None
    ):
        """Flask application initialization.

        Initialize:

        * The CLI commands.

        * Initialize the logger (Default: `app.debug`).

        * Initialize the default admin object link endpoint.
            (Default: `{"rec": "recordmetadata.details_view"}` if
            `invenio-records` is installed, otherwise `{}`).

        * Register the `pid_exists` template filter.

        * Initialize extension state.

        :param app: The Flask application
        :param minters_entry_point_group: The minters entry point group
            (Default: None).
        :param fetchers_entry_point_group: The fetchers entry point group
            (Default: None).
        :returns: PIDStore state application.
        """
        self.init_config(app)
        # Initialize CLI
        app.cli.add_command(cmd)

        # Initialize logger
        app.config.setdefault("PIDSTORE_APP_LOGGER_HANDLERS", app.debug)
        if app.config["PIDSTORE_APP_LOGGER_HANDLERS"]:
            for handler in app.logger.handlers:
                logger.addHandler(handler)

        # Initialize admin object link endpoints.
        try:
            importlib_metadata.version("invenio-records")
            app.config.setdefault(
                "PIDSTORE_OBJECT_ENDPOINTS",
                dict(
                    rec="recordmetadata.details_view",
                ),
            )
        except importlib_metadata.PackageNotFoundError:
            app.config.setdefault("PIDSTORE_OBJECT_ENDPOINTS", {})

        # Register template filter
        app.jinja_env.filters["pid_exists"] = pid_exists

        # Initialize extension state.
        state = _PIDStoreState(
            app=app,
            minters_entry_point_group=minters_entry_point_group,
            fetchers_entry_point_group=fetchers_entry_point_group,
        )
        app.extensions["invenio-pidstore"] = state
        return state

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith("PIDSTORE_") and k not in (
                "PIDSTORE_OBJECT_ENDPOINTS",
                "PIDSTORE_APP_LOGGER_HANDLERS",
            ):
                app.config.setdefault(k, getattr(config, k))

    def __getattr__(self, name):
        """Proxy to state object."""
        return getattr(self._state, name, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
# Copyright (C) 2019 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""DataCite PID provider."""

from __future__ import absolute_import

from datacite import DataCiteMDSClient
from datacite.errors import (
    DataCiteError,
    DataCiteGoneError,
    DataCiteNoContentError,
    DataCiteNotFoundError,
    HttpError,
)
from flask import current_app

from ..models import PIDStatus, logger
from .base import BaseProvider


class DataCiteProvider(BaseProvider):
    """DOI provider using DataCite API."""

    pid_type = "doi"
    """Default persistent identifier type."""

    pid_provider = "datacite"
    """Persistent identifier provider name."""

    default_status = PIDStatus.NEW
    """Default status for newly created PIDs by this provider."""

    @classmethod
    def create(cls, pid_value, **kwargs):
        """Create a new record identifier.

        For more information about parameters,
        see :meth:`invenio_pidstore.providers.base.BaseProvider.create`.

        :param pid_value: Persistent identifier value.
        :params ``**kwargs``: See
            :meth:`invenio_pidstore.providers.base.BaseProvider.create` extra
            parameters.
        :returns: A
            :class:`invenio_pidstore.providers.datacite.DataCiteProvider`
            instance.
        """
        return super(DataCiteProvider, cls).create(pid_value=pid_value, **kwargs)

    def __init__(self, pid, client=None, **kwargs):
        """Initialize provider.

        To use the default client, just configure the following variables:

        * `PIDSTORE_DATACITE_USERNAME` as username.

        * `PIDSTORE_DATACITE_PASSWORD` as password.

        * `PIDSTORE_DATACITE_DOI_PREFIX` as DOI prefix.

        * `PIDSTORE_DATACITE_TESTMODE` to `True` if it configured in test mode.

        * `PIDSTORE_DATACITE_URL` as DataCite URL.

        :param pid: A :class:`invenio_pidstore.models.PersistentIdentifier`
            instance.
        :param client: A client to access to DataCite.
            (Default: :class:`datacite.DataCiteMDSClient` instance)
        """
        super(DataCiteProvider, self).__init__(pid)
        if client is not None:
            self.api = client
        else:
            self.api = DataCiteMDSClient(
                username=current_app.config.get("PIDSTORE_DATACITE_USERNAME"),
                password=current_app.config.get("PIDSTORE_DATACITE_PASSWORD"),
                prefix=current_app.config.get("PIDSTORE_DATACITE_DOI_PREFIX"),
                test_mode=current_app.config.get("PIDSTORE_DATACITE_TESTMODE", False),
                url=current_app.config.get("PIDSTORE_DATACITE_URL"),
            )

    def reserve(self, doc):
        """Reserve a DOI (amounts to upload metadata, but not to mint).

        :param doc: Set metadata for DOI.
        :returns: `True` if is reserved successfully.
        """
        # Only registered PIDs can be updated.
        try:
            self.pid.reserve()
            self.api.metadata_post(doc)
        except (DataCiteError, HttpError):
            logger.exception("Failed to reserve in DataCite", extra=dict(pid=self.pid))
            raise
        logger.info("Successfully reserved in DataCite", extra=dict(pid=self.pid))
        return True

    def register(self, url, doc):
        """Register a DOI via the DataCite API.

        :param url: Specify the URL for the API.
        :param doc: Set metadata for DOI.
        :returns: `True` if is registered successfully.
        """
        try:
            self.pid.register()
            # Set metadata for DOI
            self.api.metadata_post(doc)
            # Mint DOI
            self.api.doi_post(self.pid.pid_value, url)
        except (DataCiteError, HttpError):
            logger.exception("Failed to register in DataCite", extra=dict(pid=self.pid))
            raise
        logger.info("Successfully registered in DataCite", extra=dict(pid=self.pid))
        return True

    def update(self, url, doc):
        """Update metadata associated with a DOI.

        This can be called before/after a DOI is registered.

        :param doc: Set metadata for DOI.
        :returns: `True` if is updated successfully.
        """
        if self.pid.is_deleted():
            logger.info("Reactivate in DataCite", extra=dict(pid=self.pid))

        try:
            # Set metadata
            self.api.metadata_post(doc)
            self.api.doi_post(self.pid.pid_value, url)
        except (DataCiteError, HttpError):
            logger.exception("Failed to update in DataCite", extra=dict(pid=self.pid))
            raise

        if self.pid.is_deleted():
            self.pid.sync_status(PIDStatus.REGISTERED)
        logger.info("Successfully updated in DataCite", extra=dict(pid=self.pid))
        return True

    def delete(self):
        """Delete a registered DOI.

        If the PID is new then it's deleted only locally.
        Otherwise, also it's deleted also remotely.

        :returns: `True` if is deleted successfully.
        """
        try:
            if self.pid.is_new():
                self.pid.delete()
            else:
                self.pid.delete()
                self.api.metadata_delete(self.pid.pid_value)
        except (DataCiteError, HttpError):
            logger.exception("Failed to delete in DataCite", extra=dict(pid=self.pid))
            raise
        logger.info("Successfully deleted in DataCite", extra=dict(pid=self.pid))
        return True

    def sync_status(self):
        """Synchronize DOI status DataCite MDS.

        :returns: `True` if is sync successfully.
        """
        status = None

        try:
            try:
                self.api.doi_get(self.pid.pid_value)
                status = PIDStatus.REGISTERED
            except DataCiteGoneError:
                status = PIDStatus.DELETED
            except DataCiteNoContentError:
                status = PIDStatus.REGISTERED
            except DataCiteNotFoundError:
                pass

            if status is None:
                try:
                    self.api.metadata_get(self.pid.pid_value)
                    status = PIDStatus.RESERVED
                except DataCiteGoneError:
                    status = PIDStatus.DELETED
                except DataCiteNoContentError:
                    status = PIDStatus.REGISTERED
                except DataCiteNotFoundError:
                    pass
        except (DataCiteError, HttpError):
            logger.exception(
                "Failed to sync status from DataCite", extra=dict(pid=self.pid)
            )
            raise

        if status is None:
            status = PIDStatus.NEW

        self.pid.sync_status(status)

        logger.info(
            "Successfully synced status from DataCite", extra=dict(pid=self.pid)
        )
        return True



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Minter tests."""

from __future__ import absolute_import, print_function

import uuid

import pytest

from invenio_pidstore import current_pidstore
from invenio_pidstore.minters import recid_minter, recid_minter_v2


def test_recid_minter(app, db):
    """Test legacy recid minter."""
    with app.app_context():
        rec_uuid = uuid.uuid4()
        data = {}

        pid = recid_minter(rec_uuid, data)

        assert pid
        assert data[app.config["PIDSTORE_RECID_FIELD"]] == pid.pid_value
        assert pid.object_type == "rec"
        assert pid.object_uuid == rec_uuid


def test_recid_minter_v2(app, db):
    """Test recommended recid minter."""
    with app.app_context():
        rec_uuid = uuid.uuid4()
        data = {}
        recid_field = app.config["PIDSTORE_RECID_FIELD"]

        pid = recid_minter_v2(rec_uuid, data)

        assert pid
        assert data[recid_field] == pid.pid_value
        assert pid.object_type == "rec"
        assert pid.object_uuid == rec_uuid

        with pytest.raises(AssertionError):
            recid_minter_v2(rec_uuid, {recid_field: "1"})


def test_register_minter(app):
    """Test base provider."""
    with app.app_context():
        current_pidstore.register_minter("anothername", recid_minter)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
# Copyright (C) 2019 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Provider tests."""

from __future__ import absolute_import, print_function

import uuid

import pytest
from datacite.errors import (
    DataCiteError,
    DataCiteGoneError,
    DataCiteNoContentError,
    DataCiteNotFoundError,
    HttpError,
)
from mock import MagicMock, patch

from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.base import BaseProvider
from invenio_pidstore.providers.datacite import DataCiteProvider
from invenio_pidstore.providers.recordid import RecordIdProvider
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2


def test_base_provider(app, db):
    """Test base provider."""
    with app.app_context():
        provider = BaseProvider.create(pid_type="test", pid_value="test")
        assert provider.pid
        assert provider.pid.pid_type == "test"
        assert provider.pid.pid_value == "test"
        assert provider.pid.pid_provider is None
        assert provider.pid.status == PIDStatus.NEW
        assert provider.pid.object_type is None
        assert provider.pid.object_uuid is None

        provider.reserve()
        assert provider.pid.is_reserved()

        provider.register()
        assert provider.pid.is_registered()

        provider.delete()
        assert provider.pid.is_deleted()

        provider.sync_status()
        provider.update()

    class TestProvider(BaseProvider):
        pid_type = "t"
        pid_provider = "testpr"
        default_status = PIDStatus.RESERVED

    with app.app_context():
        provider = TestProvider.create(pid_value="test")
        assert provider.pid
        assert provider.pid.pid_type == "t"
        assert provider.pid.pid_provider == "testpr"
        assert provider.pid.pid_value == "test"
        assert provider.pid.status == PIDStatus.RESERVED

        assert TestProvider.get("test")


def test_recordid_provider(app, db):
    """Test record id provider."""
    with app.app_context():
        provider = RecordIdProvider.create()
        assert provider.pid
        assert provider.pid.pid_type == "recid"
        assert provider.pid.pid_value == "1"
        assert provider.pid.pid_provider is None
        assert provider.pid.status == PIDStatus.RESERVED
        assert provider.pid.object_type is None
        assert provider.pid.object_uuid is None

        # Assign to object immediately
        rec_uuid = uuid.uuid4()
        provider = RecordIdProvider.create(object_type="rec", object_uuid=rec_uuid)
        assert provider.pid
        assert provider.pid.pid_type == "recid"
        assert provider.pid.pid_value == "2"
        assert provider.pid.pid_provider is None
        assert provider.pid.status == PIDStatus.REGISTERED
        assert provider.pid.object_type == "rec"
        assert provider.pid.object_uuid == rec_uuid

        pytest.raises(AssertionError, RecordIdProvider.create, pid_value="3")


def test_recordid_provider_v2(app, db):
    """Test RecordIdProviderV2."""
    with app.app_context():
        provider = RecordIdProviderV2.create()
        assert provider.pid
        assert provider.pid.pid_type == "recid"
        pid_value = provider.pid.pid_value
        part1, part2 = pid_value.split("-")
        assert len(part1) == 5
        assert len(part2) == 5
        assert provider.pid.pid_provider is None
        assert provider.pid.status == PIDStatus.RESERVED
        assert provider.pid.object_type is None
        assert provider.pid.object_uuid is None

        # Assign to object immediately
        rec_uuid = uuid.uuid4()
        provider = RecordIdProviderV2.create(object_type="rec", object_uuid=rec_uuid)
        assert provider.pid
        assert provider.pid.pid_type == "recid"
        pid_value = provider.pid.pid_value
        part1, part2 = pid_value.split("-")
        assert len(part1) == 5
        assert len(part2) == 5
        assert provider.pid.pid_provider is None
        assert provider.pid.status == PIDStatus.REGISTERED
        assert provider.pid.object_type == "rec"
        assert provider.pid.object_uuid == rec_uuid

        pytest.raises(AssertionError, RecordIdProviderV2.create, pid_value="3")

        # Options
        provider = RecordIdProviderV2.create(
            options={"length": 3, "checksum": True, "split_every": 1}
        )
        pid_value = provider.pid.pid_value
        part1, part2, part3 = pid_value.split("-")
        assert len(part1) == 1
        assert len(part2) == 1
        assert len(part3) == 1


def test_datacite_create_get(app, db):
    """Test datacite provider create/get."""
    with app.app_context():
        provider = DataCiteProvider.create("10.1234/a")
        assert provider.pid.status == PIDStatus.NEW
        assert provider.pid.pid_provider == "datacite"

        # Crete passing client kwarg to provider object creation
        provider = DataCiteProvider.create("10.1234/b", client=MagicMock())
        assert provider.pid.status == PIDStatus.NEW
        assert provider.pid.pid_provider == "datacite"
        assert isinstance(provider.api, MagicMock)

        provider = DataCiteProvider.get("10.1234/a")
        assert provider.pid.status == PIDStatus.NEW
        assert provider.pid.pid_provider == "datacite"

        provider = DataCiteProvider.get("10.1234/a", client=MagicMock())
        assert isinstance(provider.api, MagicMock)


def test_datacite_reserve_register_update_delete(app, db):
    """Test datacite provider reserve."""
    with app.app_context():
        api = MagicMock()
        provider = DataCiteProvider.create("10.1234/a", client=api)
        assert provider.reserve("mydoc")
        assert provider.pid.status == PIDStatus.RESERVED
        api.metadata_post.assert_called_with("mydoc")

        assert provider.register("myurl", "anotherdoc")
        assert provider.pid.status == PIDStatus.REGISTERED
        api.metadata_post.assert_called_with("anotherdoc")
        api.doi_post.assert_called_with("10.1234/a", "myurl")

        assert provider.update("anotherurl", "yetanother")
        assert provider.pid.status == PIDStatus.REGISTERED
        api.metadata_post.assert_called_with("yetanother")
        api.doi_post.assert_called_with("10.1234/a", "anotherurl")

        assert provider.delete()
        assert provider.pid.status == PIDStatus.DELETED
        api.metadata_delete.assert_called_with("10.1234/a")

        assert provider.update("newurl", "newdoc")
        assert provider.pid.status == PIDStatus.REGISTERED
        api.metadata_post.assert_called_with("newdoc")
        api.doi_post.assert_called_with("10.1234/a", "newurl")


@patch("invenio_pidstore.providers.datacite.logger")
def test_datacite_error_reserve(logger, app, db):
    """Test reserve errors."""
    with app.app_context():
        api = MagicMock()
        provider = DataCiteProvider.create("10.1234/a", client=api)

        api.metadata_post.side_effect = DataCiteError
        pytest.raises(DataCiteError, provider.reserve, "testdoc")
        assert logger.exception.call_args[0][0] == "Failed to reserve in DataCite"


@patch("invenio_pidstore.providers.datacite.logger")
def test_datacite_error_register_update(logger, app, db):
    """Test register errors."""
    with app.app_context():
        api = MagicMock()
        provider = DataCiteProvider.create("10.1234/a", client=api)

        api.doi_post.side_effect = DataCiteError
        pytest.raises(DataCiteError, provider.register, "testurl", "testdoc")
        assert logger.exception.call_args[0][0] == "Failed to register in DataCite"

        pytest.raises(DataCiteError, provider.update, "testurl", "testdoc")
        assert logger.exception.call_args[0][0] == "Failed to update in DataCite"


@patch("invenio_pidstore.providers.datacite.logger")
def test_datacite_error_delete(logger, app, db):
    """Test reserve errors."""
    with app.app_context():
        api = MagicMock()
        provider = DataCiteProvider.create("10.1234/a", client=api)

        # DOIs in new state doesn't contact datacite
        api.metadata_delete.side_effect = DataCiteError
        assert provider.delete()

        # Already registered DOIs do contact datacite to delete
        provider = DataCiteProvider.create(
            "10.1234/b", client=api, status=PIDStatus.REGISTERED
        )

        api.metadata_delete.side_effect = DataCiteError
        pytest.raises(DataCiteError, provider.delete)
        assert logger.exception.call_args[0][0] == "Failed to delete in DataCite"


@patch("invenio_pidstore.providers.datacite.logger")
def test_datacite_sync(logger, app, db):
    """Test sync."""
    with app.app_context():
        api = MagicMock()

        provider = DataCiteProvider.create("10.1234/a", client=api)
        assert provider.pid.status == PIDStatus.NEW

        # Status can be set from api.doi_get reply
        assert provider.sync_status()
        assert provider.pid.status == PIDStatus.REGISTERED
        api.doi_get.assert_called_with(provider.pid.pid_value)

        api.doi_get.side_effect = DataCiteGoneError
        assert provider.sync_status()
        assert provider.pid.status == PIDStatus.DELETED

        api.doi_get.side_effect = DataCiteNoContentError
        assert provider.sync_status()
        assert provider.pid.status == PIDStatus.REGISTERED

        # Status *cannot/ be set from api.doi_get reply
        # Try with api.metadata_get
        api.doi_get.side_effect = DataCiteNotFoundError
        assert provider.sync_status()
        assert provider.pid.status == PIDStatus.RESERVED
        api.metadata_get.assert_called_with(provider.pid.pid_value)

        api.doi_get.side_effect = DataCiteNotFoundError
        api.metadata_get.side_effect = DataCiteGoneError
        assert provider.sync_status()
        assert provider.pid.status == PIDStatus.DELETED

        api.doi_get.side_effect = DataCiteNotFoundError
        api.metadata_get.side_effect = DataCiteNoContentError
        assert provider.sync_status()
        assert provider.pid.status == PIDStatus.REGISTERED

        api.doi_get.side_effect = DataCiteNotFoundError
        api.metadata_get.side_effect = DataCiteNotFoundError
        assert provider.sync_status()
        assert provider.pid.status == PIDStatus.NEW

        api.doi_get.side_effect = HttpError
        assert provider.pid.status == PIDStatus.NEW
        pytest.raises(HttpError, provider.sync_status)
        assert provider.pid.status == PIDStatus.NEW
        assert logger.exception.call_args[0][0] == "Failed to sync status from DataCite"


@patch("invenio_pidstore.providers.recordid_v2.base32")
def test_recordidv2_generate_id_calls_base32_generate(patched_base32, app):
    original_options = app.config.get("PIDSTORE_RECORDID_OPTIONS")
    app.config["PIDSTORE_RECORDID_OPTIONS"] = {
        "length": 8,
        "split_every": 4,
        "checksum": False,
    }

    RecordIdProviderV2.generate_id()

    patched_base32.generate.assert_called_with(length=8, split_every=4, checksum=False)

    app.config["PIDSTORE_RECORDID_OPTIONS"] = original_options



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2023-2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Persistent identifier store and registration."""

from __future__ import absolute_import, print_function

import logging
import uuid
from enum import Enum

import six
from invenio_db import db
from invenio_i18n import lazy_gettext as _
from sqlalchemy import func
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy_utils.models import Timestamp
from sqlalchemy_utils.types import ChoiceType, UUIDType

from .errors import (
    PIDAlreadyExists,
    PIDDoesNotExistError,
    PIDInvalidAction,
    PIDObjectAlreadyAssigned,
)

logger = logging.getLogger("invenio-pidstore")


PID_STATUS_TITLES = {
    "NEW": _("New"),
    "RESERVED": _("Reserved"),
    "REGISTERED": _("Registered"),
    "REDIRECTED": _("Redirected"),
    "DELETED": _("Deleted"),
}


class PIDStatus(Enum):
    """Constants for possible status of any given PID."""

    __order__ = "NEW RESERVED REGISTERED REDIRECTED DELETED"

    NEW = "N"
    """PID has *not* yet been registered with the service provider."""

    RESERVED = "K"
    """PID reserved in the service provider but not yet fully registered."""

    REGISTERED = "R"
    """PID has been registered with the service provider."""

    REDIRECTED = "M"
    """PID has been redirected to another persistent identifier."""

    DELETED = "D"
    """PID has been deleted/inactivated with the service provider.

    This should happen very rarely, and must be kept track of, as the PID
    should not be reused for something else.
    """

    def __init__(self, value):
        """Hack."""

    def __eq__(self, other):
        """Equality test."""
        return self.value == other

    def __str__(self):
        """Return its value."""
        return self.value

    @property
    def title(self):
        """Return human readable title."""
        return PID_STATUS_TITLES[self.name]


class PersistentIdentifier(db.Model, Timestamp):
    """Store and register persistent identifiers.

    Assumptions:
      * Persistent identifiers can be represented as a string of max 255 chars.
      * An object has many persistent identifiers.
      * A persistent identifier has one and only one object.
    """

    __tablename__ = "pidstore_pid"
    __table_args__ = (
        db.Index("uidx_type_pid", "pid_type", "pid_value", unique=True),
        db.Index("idx_status", "status"),
        db.Index("idx_object", "object_type", "object_uuid"),
    )

    id = db.Column(db.Integer, primary_key=True)
    """Id of persistent identifier entry."""

    pid_type = db.Column(db.String(6), nullable=False)
    """Persistent Identifier Schema."""

    pid_value = db.Column(db.String(255), nullable=False)
    """Persistent Identifier."""

    pid_provider = db.Column(db.String(8), nullable=True)
    """Persistent Identifier Provider"""

    status = db.Column(ChoiceType(PIDStatus, impl=db.CHAR(1)), nullable=False)
    """Status of persistent identifier, e.g. registered, reserved, deleted."""

    object_type = db.Column(db.String(3), nullable=True)
    """Object Type - e.g. rec for record."""

    object_uuid = db.Column(UUIDType, nullable=True)
    """Object ID - e.g. a record id."""

    #
    # Class methods
    #
    @classmethod
    def create(
        cls,
        pid_type,
        pid_value,
        pid_provider=None,
        status=PIDStatus.NEW,
        object_type=None,
        object_uuid=None,
    ):
        """Create a new persistent identifier with specific type and value.

        :param pid_type: Persistent identifier type.
        :param pid_value: Persistent identifier value.
        :param pid_provider: Persistent identifier provider. (default: None).
        :param status: Current PID status.
            (Default: :attr:`invenio_pidstore.models.PIDStatus.NEW`)
        :param object_type: The object type is a string that identify its type.
            (default: None).
        :param object_uuid: The object UUID. (default: None).
        :returns: A :class:`invenio_pidstore.models.PersistentIdentifier`
            instance.
        """
        try:
            with db.session.begin_nested():
                obj = cls(
                    pid_type=pid_type,
                    pid_value=pid_value,
                    pid_provider=pid_provider,
                    status=status,
                )
                if object_type and object_uuid:
                    obj.assign(object_type, object_uuid)
                db.session.add(obj)
            logger.info(
                "Created PID {0}:{1}".format(pid_type, pid_value), extra={"pid": obj}
            )
        except IntegrityError:
            logger.exception(
                "PID already exists: %s:%s",
                pid_type,
                pid_value,
                extra=dict(
                    pid_type=pid_type,
                    pid_value=pid_value,
                    pid_provider=pid_provider,
                    status=status,
                    object_type=object_type,
                    object_uuid=object_uuid,
                ),
            )
            raise PIDAlreadyExists(pid_type=pid_type, pid_value=pid_value)
        except SQLAlchemyError:
            logger.exception(
                "Failed to create PID: %s:%s",
                pid_type,
                pid_value,
                extra=dict(
                    pid_type=pid_type,
                    pid_value=pid_value,
                    pid_provider=pid_provider,
                    status=status,
                    object_type=object_type,
                    object_uuid=object_uuid,
                ),
            )
            raise
        return obj

    @classmethod
    def get(cls, pid_type, pid_value, pid_provider=None):
        """Get persistent identifier.

        :param pid_type: Persistent identifier type.
        :param pid_value: Persistent identifier value.
        :param pid_provider: Persistent identifier provider. (default: None).
        :raises: :exc:`invenio_pidstore.errors.PIDDoesNotExistError` if no
            PID is found.
        :returns: A :class:`invenio_pidstore.models.PersistentIdentifier`
            instance.
        """
        try:
            args = dict(pid_type=pid_type, pid_value=six.text_type(pid_value))
            if pid_provider:
                args["pid_provider"] = pid_provider
            return db.session.query(cls).filter_by(**args).one()
        except NoResultFound:
            raise PIDDoesNotExistError(pid_type, pid_value)

    @classmethod
    def get_by_object(cls, pid_type, object_type, object_uuid):
        """Get a persistent identifier for a given object.

        :param pid_type: Persistent identifier type.
        :param object_type: The object type is a string that identify its type.
        :param object_uuid: The object UUID.
        :raises invenio_pidstore.errors.PIDDoesNotExistError: If no PID is
            found.
        :returns: A :class:`invenio_pidstore.models.PersistentIdentifier`
            instance.
        """
        try:
            return (
                db.session.query(cls)
                .filter_by(
                    pid_type=pid_type, object_type=object_type, object_uuid=object_uuid
                )
                .one()
            )
        except NoResultFound:
            raise PIDDoesNotExistError(pid_type, None)

    #
    # Assigned object methods
    #
    def has_object(self):
        """Determine if this PID has an assigned object.

        :returns: `True` if the PID has a object assigned.
        """
        return bool(self.object_type and self.object_uuid)

    def get_assigned_object(self, object_type=None):
        """Return the current assigned object UUID.

        :param object_type: If it's specified, returns only if the PID
            object_type is the same, otherwise returns None. (default: None).
        :returns: The object UUID.
        """
        if object_type is not None:
            if self.object_type == object_type:
                return self.object_uuid
            else:
                return None
        return self.object_uuid

    def assign(self, object_type, object_uuid, overwrite=False):
        """Assign this persistent identifier to a given object.

        Note, the persistent identifier must first have been reserved. Also,
        if an existing object is already assigned to the pid, it will raise an
        exception unless overwrite=True.

        :param object_type: The object type is a string that identify its type.
        :param object_uuid: The object UUID.
        :param overwrite: Force PID overwrites in case was previously assigned.
        :raises invenio_pidstore.errors.PIDInvalidAction: If the PID was
            previously deleted.
        :raises invenio_pidstore.errors.PIDObjectAlreadyAssigned: If the PID
            was previously assigned with a different type/uuid.
        :returns: `True` if the PID is successfully assigned.
        """
        if self.is_deleted():
            raise PIDInvalidAction(
                "You cannot assign objects to a deleted/redirected persistent"
                " identifier."
            )

        if not isinstance(object_uuid, uuid.UUID):
            object_uuid = uuid.UUID(object_uuid)

        if self.object_type or self.object_uuid:
            # The object is already assigned to this pid.
            if object_type == self.object_type and object_uuid == self.object_uuid:
                return True
            if not overwrite:
                raise PIDObjectAlreadyAssigned(object_type, object_uuid)
            self.unassign()

        try:
            with db.session.begin_nested():
                self.object_type = object_type
                self.object_uuid = object_uuid
                db.session.add(self)
        except SQLAlchemyError:
            logger.exception(
                "Failed to assign %s:%s", object_type, object_uuid, extra=dict(pid=self)
            )
            raise
        logger.info(
            "Assigned object {0}:{1}".format(object_type, object_uuid),
            extra=dict(pid=self),
        )
        return True

    def unassign(self):
        """Unassign the registered object.

        Note:
        Only registered PIDs can be redirected so we set it back to registered.

        :returns: `True` if the PID is successfully unassigned.

        """
        if self.object_uuid is None and self.object_type is None:
            return True

        try:
            with db.session.begin_nested():
                if self.is_redirected():
                    db.session.delete(db.session.get(Redirect, self.object_uuid))
                    # Only registered PIDs can be redirected so we set it back
                    # to registered
                    self.status = PIDStatus.REGISTERED
                self.object_type = None
                self.object_uuid = None
                db.session.add(self)
        except SQLAlchemyError:
            logger.exception("Failed to unassign object.", extra=dict(pid=self))
            raise
        logger.info("Unassigned object from {0}.".format(self), extra=dict(pid=self))
        return True

    def get_redirect(self):
        """Get redirected persistent identifier.

        :returns: The :class:`invenio_pidstore.models.PersistentIdentifier`
            instance.
        """
        return db.session.get(Redirect, self.object_uuid).pid

    #
    # Status methods.
    #
    def redirect(self, pid):
        """Redirect persistent identifier to another persistent identifier.

        :param pid: The :class:`invenio_pidstore.models.PersistentIdentifier`
            where redirect the PID.
        :raises invenio_pidstore.errors.PIDInvalidAction: If the PID is not
            registered or is not already redirecting to another PID.
        :raises invenio_pidstore.errors.PIDDoesNotExistError: If PID is not
            found.
        :returns: `True` if the PID is successfully redirect.
        """
        if not (self.is_registered() or self.is_redirected()):
            raise PIDInvalidAction("Persistent identifier is not registered.")

        try:
            with db.session.begin_nested():
                if self.is_redirected():
                    r = db.session.get(Redirect, self.object_uuid)
                    r.pid = pid
                else:
                    with db.session.begin_nested():
                        r = Redirect(pid=pid)
                        db.session.add(r)

                self.status = PIDStatus.REDIRECTED
                self.object_type = None
                self.object_uuid = r.id
                db.session.add(self)
        except IntegrityError:
            raise PIDDoesNotExistError(pid.pid_type, pid.pid_value)
        except SQLAlchemyError:
            logger.exception("Failed to redirect to %s", pid, extra=dict(pid=self))
            raise
        logger.info("Redirected PID to {0}".format(pid), extra=dict(pid=self))
        return True

    def reserve(self):
        """Reserve the persistent identifier.

        Note, the reserve method may be called multiple times, even if it was
        already reserved.

        :raises: :exc:`invenio_pidstore.errors.PIDInvalidAction` if the PID is
            not new or is not already reserved a PID.
        :returns: `True` if the PID is successfully reserved.
        """
        if not (self.is_new() or self.is_reserved()):
            raise PIDInvalidAction("Persistent identifier is not new or reserved.")

        try:
            with db.session.begin_nested():
                self.status = PIDStatus.RESERVED
                db.session.add(self)
        except SQLAlchemyError:
            logger.exception("Failed to reserve PID.", extra=dict(pid=self))
            raise
        logger.info("Reserved PID.", extra=dict(pid=self))
        return True

    def register(self):
        """Register the persistent identifier with the provider.

        :raises invenio_pidstore.errors.PIDInvalidAction: If the PID is not
            already registered or is deleted or is a redirection to another
            PID.
        :returns: `True` if the PID is successfully register.
        """
        if self.is_registered() or self.is_deleted() or self.is_redirected():
            raise PIDInvalidAction(
                "Persistent identifier has already been registered" " or is deleted."
            )

        try:
            with db.session.begin_nested():
                self.status = PIDStatus.REGISTERED
                db.session.add(self)
        except SQLAlchemyError:
            logger.exception("Failed to register PID.", extra=dict(pid=self))
            raise
        logger.info("Registered PID.", extra=dict(pid=self))
        return True

    def delete(self):
        """Delete the persistent identifier.

        If the persistent identifier haven't been registered yet, it is
        removed from the database. Otherwise, it's marked as
        :attr:`invenio_pidstore.models.PIDStatus.DELETED`.

        :returns: `True` if the PID is successfully removed.
        """
        removed = False
        try:
            with db.session.begin_nested():
                if self.is_new():
                    # New persistent identifier which haven't been registered
                    # yet.
                    db.session.delete(self)
                    removed = True
                else:
                    self.status = PIDStatus.DELETED
                    db.session.add(self)
        except SQLAlchemyError:
            logger.exception("Failed to delete PID.", extra=dict(pid=self))
            raise

        if removed:
            logger.info("Deleted PID (removed).", extra=dict(pid=self))
        else:
            logger.info("Deleted PID.", extra=dict(pid=self))
        return True

    def sync_status(self, status):
        """Synchronize persistent identifier status.

        Used when the provider uses an external service, which might have been
        modified outside of our system.

        :param status: The new status to set.
        :returns: `True` if the PID is successfully sync.
        """
        if self.status == status:
            return True

        try:
            with db.session.begin_nested():
                self.status = status
                db.session.add(self)
        except SQLAlchemyError:
            logger.exception("Failed to sync status %s.", status, extra=dict(pid=self))
            raise
        logger.info("Synced PID status to {0}.".format(status), extra=dict(pid=self))
        return True

    def is_redirected(self):
        """Return true if the persistent identifier has been registered."""
        return self.status == PIDStatus.REDIRECTED

    def is_registered(self):
        """Return true if the persistent identifier has been registered.

        :returns: A :class:`invenio_pidstore.models.PIDStatus` status.
        """
        return self.status == PIDStatus.REGISTERED

    def is_deleted(self):
        """Return true if the persistent identifier has been deleted.

        :returns: A boolean value.
        """
        return self.status == PIDStatus.DELETED

    def is_new(self):
        """Return true if the PID is new.

        :returns: A boolean value.
        """
        return self.status == PIDStatus.NEW

    def is_reserved(self):
        """Return true if the PID has been reserved.

        :returns: A boolean value.
        """
        return self.status == PIDStatus.RESERVED

    def __repr__(self):
        """Get representation of object."""
        return "<PersistentIdentifier {0}:{1}{3} ({2})>".format(
            self.pid_type,
            self.pid_value,
            self.status,
            (
                " / {0}:{1}".format(self.object_type, self.object_uuid)
                if self.object_type
                else ""
            ),
        )


class Redirect(db.Model, Timestamp):
    """Redirect for a persistent identifier.

    You can redirect a PID to another one.

    E.g.

    .. code-block:: python

        pid1 = PersistentIdentifier.get(pid_type="recid", pid_value="1")
        pid2 = PersistentIdentifier.get(pid_type="recid", pid_value="2")
        pid1.redirect(pid=pid2)
        assert pid2.pid_value == pid.get_redirect().pid_value
    """

    __tablename__ = "pidstore_redirect"
    id = db.Column(UUIDType, default=uuid.uuid4, primary_key=True)
    """Id of redirect entry."""

    pid_id = db.Column(
        db.Integer,
        db.ForeignKey(PersistentIdentifier.id, onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
    )
    """Persistent identifier."""

    pid = db.relationship(PersistentIdentifier, backref="redirects")
    """Relationship to persistent identifier."""


class RecordIdentifier(db.Model):
    """Sequence generator for integer record identifiers.

    The sole purpose of this model is to generate integer record identifiers in
    sequence using the underlying database's auto increment features in a
    transaction friendly manner. The feature is primarily provided to support
    legacy Invenio instances to continue their current record identifier
    scheme. For new instances we strong encourage to not use auto incrementing
    record identifiers, but instead use e.g. UUIDs as record identifiers.
    """

    __tablename__ = "pidstore_recid"

    recid = db.Column(
        db.BigInteger().with_variant(db.Integer, "sqlite"),
        primary_key=True,
        autoincrement=True,
    )

    @classmethod
    def next(cls):
        """Return next available record identifier."""
        try:
            with db.session.begin_nested():
                obj = cls()
                db.session.add(obj)
        except IntegrityError:  # pragma: no cover
            with db.session.begin_nested():
                # Someone has likely modified the table without using the
                # models API. Let's fix the problem.
                cls._set_sequence(cls.max())
                obj = cls()
                db.session.add(obj)
        return obj.recid

    @classmethod
    def max(cls):
        """Get max record identifier."""
        max_recid = db.session.query(func.max(cls.recid)).scalar()
        return max_recid if max_recid else 0

    @classmethod
    def _set_sequence(cls, val):
        """Internal function to reset sequence to specific value.

        Note: this function is for PostgreSQL compatibility.

        :param val: The value to be set.
        """
        if db.engine.dialect.name == "postgresql":  # pragma: no cover
            db.session.execute(
                "SELECT setval(pg_get_serial_sequence("
                "'{0}', 'recid'), :newval)".format(cls.__tablename__),
                dict(newval=val),
            )

    @classmethod
    def insert(cls, val):
        """Insert a record identifier.

        :param val: The `recid` column value to insert.
        """
        with db.session.begin_nested():
            obj = cls(recid=val)
            db.session.add(obj)
            cls._set_sequence(cls.max())


__all__ = (
    "PersistentIdentifier",
    "PIDStatus",
    "RecordIdentifier",
    "Redirect",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Admin interface tests."""

from __future__ import absolute_import, print_function

import uuid

from flask_admin import Admin, menu
from invenio_db import db

from invenio_pidstore.admin import FilterUUID, object_formatter, pid_adminview
from invenio_pidstore.models import PersistentIdentifier


def test_admin(app):
    """Test flask-admin interace."""
    admin = Admin(app, name="AdminExt")

    pid_kwargs = dict(pid_adminview)

    assert "model" in pid_adminview
    assert "modelview" in pid_adminview

    # Register both models in admin
    pid_model = pid_kwargs.pop("model")
    pid_mv = pid_kwargs.pop("modelview")
    admin.add_view(pid_mv(pid_model, db.session, **pid_kwargs))

    # Check if generated admin menu contains the correct items
    menu_items = {str(item.name): item for item in admin.menu()}

    # PIDStore should be a category
    assert "Records" in menu_items
    assert menu_items["Records"].is_category()
    assert isinstance(menu_items["Records"], menu.MenuCategory)

    # Items in PIDStore menu should be the modelviews
    submenu_items = {
        str(item.name): item for item in menu_items["Records"].get_children()
    }
    assert "Persistent Identifier" in submenu_items
    assert isinstance(submenu_items["Persistent Identifier"], menu.MenuView)


def test_filter_uuid(app, db):
    """Test FilterUUID."""
    with app.app_context():
        myuuid = uuid.uuid4()
        PersistentIdentifier.create(
            "doi", "10.1234/a", object_type="tst", object_uuid=myuuid
        )

        query = FilterUUID(PersistentIdentifier.object_uuid, "Test").apply(
            PersistentIdentifier.query, str(myuuid), None
        )
        assert query.count() == 1


def test_object_formatter(app, db):
    """Test FilterUUID."""

    @app.route("/<id>")
    def test_detail(id=None):
        return str(id)

    with app.test_request_context():
        app.config["PIDSTORE_OBJECT_ENDPOINTS"]["tst"] = "test_detail"
        pid = PersistentIdentifier.create(
            "doi", "10.1234/a", object_type="tst", object_uuid=uuid.uuid4()
        )
        assert "View" in object_formatter(None, None, pid, None)

        pid = PersistentIdentifier.create(
            "doi",
            "10.1234/b",
        )
        assert object_formatter(None, None, pid, None) == ""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create pidstore tables."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "999c62899c20"
down_revision = "f615cee99600"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        "pidstore_pid",
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("pid_type", sa.String(length=6), nullable=False),
        sa.Column("pid_value", sa.String(length=255), nullable=False),
        sa.Column("pid_provider", sa.String(length=8), nullable=True),
        sa.Column("status", sa.CHAR(1), nullable=False),
        sa.Column("object_type", sa.String(length=3), nullable=True),
        sa.Column("object_uuid", sqlalchemy_utils.types.uuid.UUIDType(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "idx_object", "pidstore_pid", ["object_type", "object_uuid"], unique=False
    )
    op.create_index("idx_status", "pidstore_pid", ["status"], unique=False)
    op.create_index(
        "uidx_type_pid", "pidstore_pid", ["pid_type", "pid_value"], unique=True
    )
    op.create_table(
        "pidstore_recid",
        sa.Column("recid", sa.BigInteger(), nullable=False),
        sa.PrimaryKeyConstraint("recid"),
    )
    op.create_table(
        "pidstore_redirect",
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("pid_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["pid_id"], ["pidstore_pid.id"], onupdate="CASCADE", ondelete="RESTRICT"
        ),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    """Downgrade database."""
    op.drop_table("pidstore_redirect")
    op.drop_table("pidstore_recid")
    op.drop_index("uidx_type_pid", table_name="pidstore_pid")
    op.drop_index("idx_status", table_name="pidstore_pid")
    op.drop_index("idx_object", table_name="pidstore_pid")
    op.drop_table("pidstore_pid")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""CLI tests."""

from __future__ import absolute_import, print_function

import uuid

from click.testing import CliRunner
from flask.cli import ScriptInfo

from invenio_pidstore.cli import pid as cmd
from invenio_pidstore.models import PersistentIdentifier, PIDStatus


def test_pid_creation(app, db):
    """Test pid creation."""
    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda: app)

    with runner.isolated_filesystem():
        with app.app_context():
            assert PersistentIdentifier.query.count() == 0

        result = runner.invoke(cmd, ["create", "doi", "10.1234/foo"], obj=script_info)
        assert 0 == result.exit_code

        with app.app_context():
            assert PersistentIdentifier.query.count() == 1
            pid = PersistentIdentifier.get("doi", "10.1234/foo")
            assert pid.pid_type == "doi"
            assert pid.pid_value == "10.1234/foo"
            assert pid.pid_provider is None
            assert pid.status == PIDStatus.NEW
            assert pid.object_type is None
            assert pid.object_uuid is None

        rec_uuid = uuid.uuid4()

        # Bad parameter status:
        result = runner.invoke(
            cmd,
            [
                "create",
                "recid",
                "2",
                "--status",
                "BADPARAMETER",
                "--type",
                "rec",
                "--uuid",
                str(rec_uuid),
            ],
            obj=script_info,
        )
        assert 2 == result.exit_code

        # Any or both type and uuid must be defined:
        result = runner.invoke(
            cmd,
            [
                "create",
                "recid",
                "2",
                "--type",
                "rec",
            ],
            obj=script_info,
        )
        assert 2 == result.exit_code

        result = runner.invoke(
            cmd,
            [
                "create",
                "recid",
                "2",
                "--uuid",
                str(rec_uuid),
            ],
            obj=script_info,
        )
        assert 2 == result.exit_code

        # Everything should be fine now:
        result = runner.invoke(
            cmd,
            [
                "create",
                "recid",
                "2",
                "--status",
                "REGISTERED",
                "--type",
                "rec",
                "--uuid",
                str(rec_uuid),
            ],
            obj=script_info,
        )
        assert 0 == result.exit_code

        with app.app_context():
            assert PersistentIdentifier.query.count() == 2
            pid = PersistentIdentifier.get("recid", "2")
            assert pid.pid_type == "recid"
            assert pid.pid_value == "2"
            assert pid.pid_provider is None
            assert pid.status == PIDStatus.REGISTERED
            assert pid.object_type == "rec"
            assert pid.object_uuid == rec_uuid

        # Can't duplicate existing persistent identifier
        result = runner.invoke(
            cmd,
            [
                "create",
                "recid",
                "2",
            ],
            obj=script_info,
        )
        assert 1 == result.exit_code


def test_pid_assign(app, db):
    """Test pid object assignment."""
    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda: app)

    with runner.isolated_filesystem():
        # No assigned object
        result = runner.invoke(cmd, ["create", "doi", "10.1234/foo"], obj=script_info)
        assert 0 == result.exit_code

        with app.app_context():
            pid = PersistentIdentifier.get("doi", "10.1234/foo")
            assert not pid.has_object()
            assert pid.get_assigned_object() is None
            assert pid.get_assigned_object("rec") is None

        # Assign object
        rec_uuid = uuid.uuid4()
        result = runner.invoke(
            cmd,
            ["assign", "doi", "10.1234/foo", "-t", "rec", "-i", str(rec_uuid)],
            obj=script_info,
        )
        assert 0 == result.exit_code
        with app.app_context():
            pid = PersistentIdentifier.get("doi", "10.1234/foo")
            assert pid.has_object()
            assert pid.get_assigned_object() == rec_uuid
            assert pid.get_assigned_object("rec") == rec_uuid
            assert pid.get_assigned_object("oth") is None

        # Doesnt' raise
        result = runner.invoke(
            cmd,
            ["assign", "doi", "10.1234/foo", "-t", "rec", "-i", str(rec_uuid)],
            obj=script_info,
        )
        assert 0 == result.exit_code

        # Missing type or uuid:
        result = runner.invoke(
            cmd,
            [
                "assign",
                "doi",
                "10.1234/foo",
            ],
            obj=script_info,
        )
        assert 2 == result.exit_code

        result = runner.invoke(
            cmd,
            [
                "assign",
                "doi",
                "10.1234/foo",
                "-t",
                "rec",
            ],
            obj=script_info,
        )
        assert 2 == result.exit_code

        result = runner.invoke(
            cmd,
            [
                "assign",
                "doi",
                "10.1234/foo",
                "-i",
                str(rec_uuid),
            ],
            obj=script_info,
        )
        assert 2 == result.exit_code

        # Assign without overwrite (uuid as str and uuid)
        new_uuid = uuid.uuid4()
        result = runner.invoke(
            cmd,
            ["assign", "doi", "10.1234/foo", "-t", "rec", "-i", str(new_uuid)],
            obj=script_info,
        )

        assert 1 == result.exit_code

        # Assign with overwrite
        result = runner.invoke(
            cmd,
            [
                "assign",
                "doi",
                "10.1234/foo",
                "-s",
                "REGISTERED",
                "-t",
                "rec",
                "-i",
                str(new_uuid),
                "--overwrite",
            ],
            obj=script_info,
        )
        assert 0 == result.exit_code

        with app.app_context():
            pid = PersistentIdentifier.get("doi", "10.1234/foo")
            assert pid.has_object()
            assert pid.status == PIDStatus.REGISTERED
            assert pid.get_assigned_object() == new_uuid
            assert pid.get_assigned_object("rec") == new_uuid
            assert pid.get_assigned_object("oth") is None


def test_pid_unassign(app, db):
    """Test pid object unassignment."""
    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda: app)

    with runner.isolated_filesystem():
        rec_uuid = uuid.uuid4()
        # Assigned object
        result = runner.invoke(
            cmd,
            ["create", "recid", "101", "-t", "rec", "-i", str(rec_uuid)],
            obj=script_info,
        )
        assert 0 == result.exit_code

        result = runner.invoke(
            cmd,
            [
                "get",
                "recid",
                "101",
            ],
            obj=script_info,
        )
        assert 0 == result.exit_code
        assert "rec {0} N\n".format(str(rec_uuid)) == result.output

        result = runner.invoke(
            cmd,
            [
                "dereference",
                "rec",
                str(rec_uuid),
            ],
            obj=script_info,
        )
        assert 0 == result.exit_code
        assert "recid 101 None\n" == result.output

        result = runner.invoke(
            cmd,
            [
                "dereference",
                "rec",
                str(rec_uuid),
                "-s",
                "NEW",
            ],
            obj=script_info,
        )
        assert 0 == result.exit_code
        assert "recid 101 None\n" == result.output

        with app.app_context():
            pid = PersistentIdentifier.get("recid", "101")
            assert pid.has_object()
            assert pid.get_assigned_object() == rec_uuid
            assert pid.get_assigned_object("rec") == rec_uuid

        # Unassign the object
        result = runner.invoke(
            cmd,
            [
                "unassign",
                "recid",
                "101",
            ],
            obj=script_info,
        )
        assert 0 == result.exit_code

        with app.app_context():
            pid = PersistentIdentifier.get("recid", "101")
            assert not pid.has_object()
            assert pid.get_assigned_object() is None
            assert pid.get_assigned_object("rec") is None



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Fetcher tests."""

from __future__ import absolute_import, print_function

import uuid

from invenio_pidstore import current_pidstore
from invenio_pidstore.fetchers import recid_fetcher, recid_fetcher_v2
from invenio_pidstore.minters import recid_minter, recid_minter_v2


def test_recid_fetcher(app, db):
    """Test legacy recid fetcher."""
    with app.app_context():
        rec_uuid = uuid.uuid4()
        data = {}
        minted_pid = recid_minter(rec_uuid, data)
        fetched_pid = recid_fetcher(rec_uuid, data)
        assert minted_pid.pid_value == fetched_pid.pid_value
        assert fetched_pid.pid_type == fetched_pid.provider.pid_type
        assert fetched_pid.pid_type == "recid"


def test_recid_fetcher_v2(app, db):
    """Test recommended recid fetcher."""
    with app.app_context():
        rec_uuid = uuid.uuid4()
        data = {}
        minted_pid = recid_minter_v2(rec_uuid, data)

        fetched_pid = recid_fetcher_v2(rec_uuid, data)

        assert minted_pid.pid_value == fetched_pid.pid_value
        assert minted_pid.pid_type == fetched_pid.pid_type
        assert fetched_pid.pid_type == "recid"
        assert fetched_pid.pid_value == minted_pid.pid_value


def test_register_fetcher(app):
    """Test base provider."""
    with app.app_context():
        current_pidstore.register_fetcher("anothername", recid_minter)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Record ID provider."""

from __future__ import absolute_import, print_function

from ..models import PIDStatus, RecordIdentifier
from .base import BaseProvider


class RecordIdProvider(BaseProvider):
    """Record identifier provider."""

    pid_type = "recid"
    """Type of persistent identifier."""

    pid_provider = None
    """Provider name.

    The provider name is not recorded in the PID since the provider does not
    provide any additional features besides creation of record ids.
    """

    default_status = PIDStatus.RESERVED
    """Record IDs are by default registered immediately.

    Default: :attr:`invenio_pidstore.models.PIDStatus.RESERVED`
    """

    @classmethod
    def create(cls, object_type=None, object_uuid=None, **kwargs):
        """Create a new record identifier.

        Note: if the object_type and object_uuid values are passed, then the
        PID status will be automatically setted to
        :attr:`invenio_pidstore.models.PIDStatus.REGISTERED`.

        :param object_type: The object type. (Default: None.)
        :param object_uuid: The object identifier. (Default: None).
        :param kwargs: You specify the pid_value.
        """
        # Request next integer in recid sequence.
        assert "pid_value" not in kwargs
        kwargs["pid_value"] = str(RecordIdentifier.next())
        kwargs.setdefault("status", cls.default_status)
        if object_type and object_uuid:
            kwargs["status"] = PIDStatus.REGISTERED
        return super(RecordIdProvider, cls).create(
            object_type=object_type, object_uuid=object_uuid, **kwargs
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create pidstore branch."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f615cee99600"
down_revision = None
branch_labels = ("invenio_pidstore",)
depends_on = "dbdbc1b19cf2"


def upgrade():
    """Upgrade database."""


def downgrade():
    """Downgrade database."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
# Copyright (C) 2019 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""DataCite PID provider."""

from __future__ import absolute_import

import copy

from base32_lib import base32
from flask import current_app

from ..models import PIDStatus
from .base import BaseProvider


class RecordIdProviderV2(BaseProvider):
    """Record identifier provider V2.

    This is the recommended record id provider.

    It generates a random alphanumeric string as opposed to an increasing
    integer (:class:`invenio_pidstore.providers.recordid.RecordIdProvider`).
    """

    pid_type = "recid"
    """Type of persistent identifier."""

    pid_provider = None
    """Provider name.

    The provider name is not recorded in the PID since the provider does not
    provide any additional features besides creation of record ids.
    """

    default_status_with_obj = PIDStatus.REGISTERED
    """Record IDs are by default registered immediately.

    Default: :attr:`invenio_pidstore.models.PIDStatus.REGISTERED`
    """

    default_status = PIDStatus.RESERVED
    """Record IDs with an object are by default reserved.

    Default: :attr:`invenio_pidstore.models.PIDStatus.RESERVED`
    """

    @classmethod
    def generate_id(cls, options=None):
        """Generate record id."""
        passed_options = options or {}
        # WHY: A new dict needs to be created to prevent side-effects
        options = copy.deepcopy(current_app.config.get("PIDSTORE_RECORDID_OPTIONS", {}))
        options.update(passed_options)
        length = options.get("length", 10)
        split_every = options.get("split_every", 0)
        checksum = options.get("checksum", True)

        return base32.generate(
            length=length, split_every=split_every, checksum=checksum
        )

    @classmethod
    def create(cls, object_type=None, object_uuid=None, options=None, **kwargs):
        """Create a new record identifier.

        Note: if the object_type and object_uuid values are passed, then the
        PID status will be automatically setted to
        :attr:`invenio_pidstore.models.PIDStatus.REGISTERED`.

        For more information about parameters,
        see :meth:`invenio_pidstore.providers.base.BaseProvider.create`.

        :param object_type: The object type. (Default: None.)
        :param object_uuid: The object identifier. (Default: None).
        :param options: ``dict`` with optional keys:
            ``"length"`` (integer), ``"split_every"`` (integer),
            ``"checksum"`` (boolean). (Default: None).
        :param kwargs: dict to hold generated pid_value and status. See
            :meth:`invenio_pidstore.providers.base.BaseProvider.create` extra
            parameters.
        :returns: A :class:`RecordIdProviderV2` instance.
        """
        assert "pid_value" not in kwargs

        kwargs["pid_value"] = cls.generate_id(options)
        kwargs.setdefault("status", cls.default_status)

        if object_type and object_uuid:
            kwargs["status"] = cls.default_status_with_obj

        return super(RecordIdProviderV2, cls).create(
            object_type=object_type, object_uuid=object_uuid, **kwargs
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Record identifier model tests."""

from __future__ import absolute_import, print_function

from invenio_pidstore.models import RecordIdentifier


def test_record_identifier(app, db):
    """Test base provider."""
    with app.app_context():
        assert RecordIdentifier.next() == 1
        assert RecordIdentifier.next() == 2
        assert RecordIdentifier.max() == 2

        # Mess up the sequence
        with db.session.begin_nested():
            obj = RecordIdentifier(recid=3)
            db.session.add(obj)

        assert RecordIdentifier.max() == 3

        # This tests a particular problem on PostgreSQL which is using
        # sequences to generate auto incrementing columns and doesn't deal
        # nicely with having values inserted in the table.
        assert RecordIdentifier.next() == 4

        RecordIdentifier.insert(10)
        assert RecordIdentifier.next() == 11
        assert RecordIdentifier.max() == 11

        RecordIdentifier.insert(7)
        assert RecordIdentifier.max() == 11
        assert RecordIdentifier.next() == 12



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Errors for persistent identifiers."""

from __future__ import absolute_import, print_function


class PersistentIdentifierError(Exception):
    """Base class for PIDStore errors."""


class PIDValueError(PersistentIdentifierError):
    """Base class for value errors."""

    def __init__(self, pid_type, pid_value, *args, **kwargs):
        """Initialize exception."""
        self.pid_type = pid_type
        self.pid_value = pid_value
        super(PIDValueError, self).__init__(*args, **kwargs)


class PIDDoesNotExistError(PIDValueError):
    """PID does not exists error."""


class PIDAlreadyExists(PIDValueError):
    """Persistent identifier already exists error."""


class ResolverError(PersistentIdentifierError):
    """Persistent identifier does not exists."""

    def __init__(self, pid, *args, **kwargs):
        """Initialize exception."""
        self.pid = pid
        super(ResolverError, self).__init__(*args, **kwargs)


class PIDDeletedError(ResolverError):
    """Persistent identifier is deleted."""

    def __init__(self, pid, record, *args, **kwargs):
        """Initialize exception."""
        self.record = record
        super(PIDDeletedError, self).__init__(pid, *args, **kwargs)


class PIDMissingObjectError(ResolverError):
    """Persistent identifier has no object."""


class PIDUnregistered(ResolverError):
    """Persistent identifier has not been registered."""


class PIDRedirectedError(ResolverError):
    """Persistent identifier is redirected to another pid."""

    def __init__(self, pid, dest_pid, *args, **kwargs):
        """Initialize exception."""
        self.destination_pid = dest_pid
        super(PIDRedirectedError, self).__init__(pid, *args, **kwargs)


class PIDObjectAlreadyAssigned(PersistentIdentifierError):
    """Persistent identifier is already assigned to another object."""


class PIDInvalidAction(PersistentIdentifierError):
    """Invalid operation on persistent identifier in current state."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Persistent identifier minters."""

from __future__ import absolute_import, print_function

from flask import current_app

from .providers.recordid import RecordIdProvider
from .providers.recordid_v2 import RecordIdProviderV2


def recid_minter_v2(record_uuid, data):
    """Mint record identifiers with RecordIDProviderV2.

    This minter is recommended to be used when creating records to get
    PersistentIdentifier with ``object_type='rec'`` and the new random
    alphanumeric `pid_value`.

    Raises ``AssertionError`` if a ``PIDSTORE_RECID_FIELD`` entry is already in
    ``data``. The minted ``pid_value`` will be stored in that field.

    :param record_uuid: The object UUID of the record.
    :param data: The record metadata.
    :returns: A fresh `invenio_pidstore.models.PersistentIdentifier` instance.
    """
    pid_field = current_app.config["PIDSTORE_RECID_FIELD"]
    assert pid_field not in data
    provider = RecordIdProviderV2.create(object_type="rec", object_uuid=record_uuid)
    data[pid_field] = provider.pid.pid_value
    return provider.pid


def recid_minter(record_uuid, data):
    """Mint record identifiers.

    This is a minter specific for records.
    With the help of
    :class:`invenio_pidstore.providers.recordid.RecordIdProvider`, it creates
    the PID instance with `rec` as predefined `object_type`.

    Procedure followed: (we will use `control_number` as value of
    `PIDSTORE_RECID_FIELD` for the simplicity of the documentation.)

    #. If a `control_number` field is already there, a `AssertionError`
    exception is raised.

    #. The provider is initialized with the help of
    :class:`invenio_pidstore.providers.recordid.RecordIdProvider`.
    It's called with default value 'rec' for `object_type` and `record_uuid`
    variable for `object_uuid`.

    #. The new `id_value` is stored inside `data` as `control_number` field.

    :param record_uuid: The record UUID.
    :param data: The record metadata.
    :returns: A fresh `invenio_pidstore.models.PersistentIdentifier` instance.
    """
    pid_field = current_app.config["PIDSTORE_RECID_FIELD"]
    assert pid_field not in data
    provider = RecordIdProvider.create(object_type="rec", object_uuid=record_uuid)
    data[pid_field] = provider.pid.pid_value
    return provider.pid



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2022 RERO.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Extension tests."""

from __future__ import absolute_import, print_function

from flask import Flask
from mock import patch

from invenio_pidstore import InvenioPIDStore
from invenio_pidstore.models import PersistentIdentifier


def test_version():
    """Test version import."""
    from invenio_pidstore import __version__

    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask("testapp")
    ext = InvenioPIDStore(app)
    assert "invenio-pidstore" in app.extensions
    ext.register_minter(
        "testminter", lambda a, b: "deadbeef-c0de-c0de-c0de-b100dc0ffee5"
    )
    ext.register_fetcher(
        "testfetcher", lambda a, b: "deadbeef-c0de-c0de-c0de-b100dc0ffee5"
    )

    app = Flask("testapp")
    ext = InvenioPIDStore()
    assert "invenio-pidstore" not in app.extensions
    ext.init_app(app)
    assert "invenio-pidstore" in app.extensions


def test_logger():
    """Test extension initialization."""
    app = Flask("testapp")
    app.config["PIDSTORE_APP_LOGGER_HANDLERS"] = True
    InvenioPIDStore(app)


def test_invenio_records():
    """Test extension initialization."""
    app = Flask("testapp")
    with patch("invenio_pidstore.ext.importlib_metadata"):
        with patch("invenio_pidstore.ext.importlib_resources"):
            InvenioPIDStore(app)
    assert app.config["PIDSTORE_OBJECT_ENDPOINTS"]


def test_template_filters(app, db):
    """Test the template filters."""
    with app.app_context():
        # Test the 'pid_exists' template filter
        pid_exists = app.jinja_env.filters["pid_exists"]
        assert not pid_exists("pid_val0", pidtype="mock_t")
        PersistentIdentifier.create("mock_t", "pid_val0")
        db.session.commit()
        assert pid_exists("pid_val0", pidtype="mock_t")
        assert not pid_exists("foo", pidtype="mock_t")
        assert not pid_exists("pid_val0", pidtype="foo")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Admin model views for PersistentIdentifier."""

import uuid

from flask import current_app, url_for
from flask_admin.contrib.sqla import ModelView
from flask_admin.contrib.sqla.filters import FilterEqual
from markupsafe import Markup

from .models import PersistentIdentifier, PIDStatus


def _(x):
    """Identity function for string extraction."""
    return x


def object_formatter(v, c, m, p):
    """Format object view link."""
    endpoint = current_app.config["PIDSTORE_OBJECT_ENDPOINTS"].get(m.object_type)

    if endpoint and m.object_uuid:
        return Markup(
            '<a href="{0}">{1}</a>'.format(
                url_for(endpoint, id=m.object_uuid), _("View")
            )
        )
    return ""


class FilterUUID(FilterEqual):
    """UUID aware filter."""

    def apply(self, query, value, alias):
        """Convert UUID."""
        return query.filter(self.column == uuid.UUID(value))


class PersistentIdentifierModelView(ModelView):
    """ModelView for the PersistentIdentifier."""

    can_create = False
    can_edit = False
    can_delete = False
    can_view_details = True
    column_display_all_relations = True
    column_list = (
        "pid_type",
        "pid_value",
        "status",
        "object_type",
        "object_uuid",
        "created",
        "updated",
        "object",
    )
    column_labels = dict(
        pid_type=_("PID Type"),
        pid_value=_("PID"),
        pid_provider=_("Provider"),
        status=_("Status"),
        object_type=_("Object Type"),
        object_uuid=_("Object UUID"),
    )
    column_filters = (
        "pid_type",
        "pid_value",
        "object_type",
        FilterUUID(PersistentIdentifier.object_uuid, _("Object UUID")),
        FilterEqual(
            PersistentIdentifier.status,
            _("Status"),
            options=[(s.value, s.title) for s in PIDStatus],
        ),
    )
    column_searchable_list = ("pid_value",)
    column_default_sort = ("updated", True)
    column_formatters = dict(object=object_formatter)
    page_size = 25


pid_adminview = dict(
    modelview=PersistentIdentifierModelView,
    model=PersistentIdentifier,
    category=_("Records"),
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolver tests."""

from __future__ import absolute_import, print_function

import uuid

import pytest

from invenio_pidstore.errors import (
    PIDDeletedError,
    PIDDoesNotExistError,
    PIDMissingObjectError,
    PIDRedirectedError,
    PIDUnregistered,
)
from invenio_pidstore.models import PersistentIdentifier, PIDStatus
from invenio_pidstore.resolver import Resolver


def test_resolver(app, db):
    """Test the class methods of PersistentIdentifier class."""
    status = [
        PIDStatus.NEW,
        PIDStatus.RESERVED,
        PIDStatus.REGISTERED,
        PIDStatus.DELETED,
    ]

    with app.app_context():
        i = 1
        rec_a = uuid.uuid4()

        # Create pids for each status with and without object
        for s in status:
            PersistentIdentifier.create("recid", i, status=s)
            i += 1
            if s != PIDStatus.DELETED:
                PersistentIdentifier.create(
                    "recid", i, status=s, object_type="rec", object_uuid=rec_a
                )
                i += 1

        # Create a DOI
        pid_doi = PersistentIdentifier.create(
            "doi",
            "10.1234/foo",
            status=PIDStatus.REGISTERED,
            object_type="rec",
            object_uuid=rec_a,
        )

        # Create redirects
        pid = PersistentIdentifier.create("recid", i, status=PIDStatus.REGISTERED)
        i += 1
        pid.redirect(PersistentIdentifier.get("recid", "2"))
        pid = PersistentIdentifier.create("recid", i, status=PIDStatus.REGISTERED)
        pid.redirect(pid_doi)
        db.session.commit()

        # Start tests
        resolver = Resolver(pid_type="recid", object_type="rec", getter=lambda x: x)

        # Resolve non-existing pid
        pytest.raises(PIDDoesNotExistError, resolver.resolve, "100")
        pytest.raises(PIDDoesNotExistError, resolver.resolve, "10.1234/foo")

        # Resolve status new
        pytest.raises(PIDUnregistered, resolver.resolve, "1")
        pytest.raises(PIDUnregistered, resolver.resolve, "2")

        # Resolve status reserved
        pytest.raises(PIDUnregistered, resolver.resolve, "3")
        pytest.raises(PIDUnregistered, resolver.resolve, "4")

        # Resolve status registered
        pytest.raises(PIDMissingObjectError, resolver.resolve, "5")
        pid, obj = resolver.resolve("6")
        assert pid and obj == rec_a

        # Resolve status deleted
        pytest.raises(PIDDeletedError, resolver.resolve, "7")

        # Resolve status redirected
        try:
            resolver.resolve("8")
            assert False
        except PIDRedirectedError as e:
            assert e.destination_pid.pid_type == "recid"
            assert e.destination_pid.pid_value == "2"

        try:
            resolver.resolve("9")
            assert False
        except PIDRedirectedError as e:
            assert e.destination_pid.pid_type == "doi"
            assert e.destination_pid.pid_value == "10.1234/foo"

        doiresolver = Resolver(pid_type="doi", object_type="rec", getter=lambda x: x)
        pytest.raises(PIDDoesNotExistError, doiresolver.resolve, "1")
        pid, obj = doiresolver.resolve("10.1234/foo")
        assert pid and obj == rec_a


def test_resolver_deleted_object(app, db):
    """Test the class methods of PersistentIdentifier class."""
    with app.app_context():
        rec_uuid = uuid.uuid4()
        records = {
            rec_uuid: {"title": "test"},
        }
        with db.session.begin_nested():
            pid = PersistentIdentifier.create(
                "recid",
                "1",
                status=PIDStatus.REGISTERED,
                object_type="rec",
                object_uuid=rec_uuid,
            )

        with db.session.begin_nested():
            pid.delete()

        resolver = Resolver(pid_type="recid", object_type="rec", getter=records.get)

        assert pytest.raises(PIDDeletedError, resolver.resolve, "1")


def test_resolver_not_registered_only(app, db):
    """Test the resolver returns a new and reserved PID when specified."""

    status = [PIDStatus.NEW, PIDStatus.RESERVED, PIDStatus.REGISTERED]

    with app.app_context():
        rec_a = uuid.uuid4()
        # Create pids for each status with and without object
        for idx, s in enumerate(status, 1):
            PersistentIdentifier.create("recid", idx * 2 - 1, status=s)
            PersistentIdentifier.create(
                "recid", idx * 2, status=s, object_type="rec", object_uuid=rec_a
            )

        db.session.commit()

        # Start tests
        resolver = Resolver(
            pid_type="recid",
            object_type="rec",
            getter=lambda x: x,
            registered_only=False,
        )

        # Resolve status new
        pytest.raises(PIDMissingObjectError, resolver.resolve, "1")
        pid, obj = resolver.resolve("2")
        assert pid and obj == rec_a

        # Resolve status reserved
        pytest.raises(PIDMissingObjectError, resolver.resolve, "3")
        pid, obj = resolver.resolve("4")
        assert pid and obj == rec_a

        # Resolve status registered
        pytest.raises(PIDMissingObjectError, resolver.resolve, "5")
        pid, obj = resolver.resolve("6")
        assert pid and obj == rec_a



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Internal resolver for persistent identifiers."""

from __future__ import absolute_import, print_function

from sqlalchemy.orm.exc import NoResultFound

from .errors import (
    PIDDeletedError,
    PIDMissingObjectError,
    PIDRedirectedError,
    PIDUnregistered,
)
from .models import PersistentIdentifier


class Resolver(object):
    """Persistent identifier resolver.

    Helper class for retrieving an internal object for a given persistent
    identifier.
    """

    def __init__(
        self, pid_type=None, object_type=None, getter=None, registered_only=True
    ):
        """Initialize resolver.

        :param pid_type: Persistent identifier type.
        :param object_type: Object type.
        :param getter: Callable that will take an object id for the given
            object type and retrieve the internal object.
        """
        self.pid_type = pid_type
        self.object_type = object_type
        self.object_getter = getter
        self.registered_only = registered_only

    def resolve(self, pid_value):
        """Resolve a persistent identifier to an internal object.

        :param pid_value: Persistent identifier.
        :returns: A tuple containing (pid, object).
        """
        pid = PersistentIdentifier.get(self.pid_type, pid_value)

        if pid.is_new() or pid.is_reserved():
            if self.registered_only:
                raise PIDUnregistered(pid)
            else:
                obj_id = pid.get_assigned_object(object_type=self.object_type)

        if pid.is_deleted():
            obj_id = pid.get_assigned_object(object_type=self.object_type)
            try:
                obj = self.object_getter(obj_id) if obj_id else None
            except NoResultFound:
                obj = None
            raise PIDDeletedError(pid, obj)

        if pid.is_redirected():
            raise PIDRedirectedError(pid, pid.get_redirect())

        obj_id = pid.get_assigned_object(object_type=self.object_type)
        if not obj_id:
            raise PIDMissingObjectError(self.pid_type, pid_value)

        return pid, self.object_getter(obj_id)
